
import os  
import pandas as pd
import numpy as np
#from scipy import signal

import pickle

import os
import multiprocessing
import time
import fastqsplitter
from pathlib import Path
from ..util.cli import (
                        opt_barcode_length,opt_barcode_start,opt_parallel_demultiplexing,opt_clipped_demultiplexing,opt_mismatch_tolerence,opt_index_tolerence,opt_demulti_overwrite,opt_fasta,opt_library,opt_fastq1,opt_fastq2)

#secondary barcode names: secondary_signiture
params = [
    # Inputs
    opt_fasta,
    opt_fastq1,
    opt_fastq2,
    opt_library,
    opt_barcode_start,
    opt_barcode_length,

    #options
    opt_parallel_demultiplexing,
    opt_clipped_demultiplexing,
    opt_mismatch_tolerence,
    opt_index_tolerence,
    opt_demulti_overwrite


]


def makes_dict_from_fastq(fpath):
        #do I make a set of the cordinates? ideally that's the only part thats different..?


        f = open(fpath)
        lines = f.readlines()
        f.close()
        reads={}
        read_lines=[]
        for x in range(len(lines)):

            l=lines[x]
            read_lines.append(l)


            if x%4==0:
                spl=l.split()
                identifier_pre=spl[0]#5/6
                values=identifier_pre.split(":")
                identifier=(values[4]+":"+values[5]+":"+values[6])

            if x%4==3:
                reads[identifier]=read_lines
                read_lines=[]
        return reads       

def reverse_compliment(sequence):
    rev_seq=sequence[::-1]
    new_str=""
    for x in range(len(rev_seq)):
        if rev_seq[x]=="A":
            new_str+="T"
        if rev_seq[x]=="G":
            new_str+="C"
        if rev_seq[x]=="C":
            new_str+="G"
        if rev_seq[x]=="T":
            new_str+="A"
    return new_str


class Sequence_Obj():
    def __init__(self,
    sequence,
    name,
    fastq1_path,
    fastq2_path,
    workspace,
    paired=True,
    fwd_primer="",
    rev_primer="",
    secondary_signature="",secondary_signature_start="",secondary_signature_end="",
    rev_secondary_signature="",rev_secondary_signature_start="",rev_secondary_signature_end="",
    barcode_start=-1,barcode_end=-1,barcode="",
    rev_barcode="",rev_barcode_start="",rev_barcode_end=""):

        """
        This object will be hold the reference objects:
            including a dictionary of the ranges and the bv/mh associated with each of those
        """

        #input_df=pd.read_csv()
        self.demutliplex_fasta=""
        self.sequence=sequence
        self.sample_name=name
        self.paired=paired
        self.sequence=sequence
        self.fwd_primer=fwd_primer
        self.rev_primer=rev_primer
        self.paired=paired
        self.workspace=workspace


        self.fastq1_path=fastq1_path
        self.fastq2_path=fastq2_path
        self.fastq_paths={1:fastq1_path,2:fastq2_path}
        #barcode info
        self.barcode_start=barcode_start
        self.barcode_end=barcode_end
        self.barcode=barcode

        self.rev_barcode=rev_barcode
        self.rev_barcode_start=rev_barcode_start
        self.rev_barcode_end=rev_barcode_end

        #secondary signiture info
        self.secondary_signature=secondary_signature
        self.secondary_signature_start=secondary_signature_start
        self.secondary_signature_end=secondary_signature_end

        self.rev_secondary_signature=rev_secondary_signature
        self.rev_secondary_signature_start=rev_secondary_signature_start
        self.rev_secondary_signature_end=rev_secondary_signature_end



        #self.sequence_folder=workspace
        self.fastqs={1:"",2:""}

        self.fastq1_init_bc=set()
        self.fastq2_init_bc=set()
        
        clip_grepped_fq1={}
        clip_grepped_fq2={}

        #reads grepped by rev barcode search
        self.fastq1_rev_bc=set()
        self.fastq2_rev_bc=set()

        rev_clip_grepped_fq1={}
        rev_clip_grepped_fq2={}

        #ids of reads filtered out due to the secondary signiture
        self.fastq1_sec_filtered=set()
        self.fastq2_sec_filtered=set()

        #ids of reads filtered out due to reverse compliment of the secondary signiture
        self.fastq1_sec_rev_filtered=set()
        self.fastq2_sec_rev_filtered=set()

        #this is the total before secondary signiture filtering
        self.fastq1_unfiltered=set()
        self.fastq2_unfiltered=set()


        #this is post filtering or if no sec_sign filtering is going to happen
        self.fastq1_filtered=set()
        self.fastq2_filtered=set()


        #these too will only really need one since theyre the same reads either way
        #set of distnct reads from either fq1 or fq2
        self.fastq_union=set()

        #set of reads found in both fastq
        self.fastq_intersection=set()

        #organized by the amount of NT clipped from barcode and then searched for
        clip_grepped_fq1={}
        clip_grepped_fq2={}

        #grepped_done_file=

class super_fastq():

    def __init__(self,fpath:str,split_count=10,fastq_name:str=None,super_dir="")-> None:
        #print(super_dir)
        if(fastq_name==None):
            fq=fpath.split("/")[-1]
            dir_name=super_dir+fq.split("_R")[0]+"_"+fq.split("_R")[-1][0]+"_pickle_split/"
        else:
            dir_name=super_dir+fastq_name
        #dir_name+=super_dir
        os.makedirs(dir_name,exist_ok=True)
        try:

            fq_size=os.path.getsize(fpath)
        except:
            raise Exception("File could not be opened, please check if the path you gave is correct!")


        split_list=[]
        pickle_file_list=[]
        self.original_fastq=fpath
        self.id_to_pickle_dict={}
        if fq_size<100000:
            self.split_count=1
        else:
            self.split_count=split_count

        for x in range(self.split_count):
            sp=dir_name + "split_"+str(x)+".fastq"
            sp_pickle=dir_name + "split_"+str(x)+".p"
            self.id_to_pickle_dict[x]=dict()

            pickle_file_list.append(sp_pickle)
            split_list.append(sp)

        self.split_list=split_list
        self.pickle_file_list=pickle_file_list
        
        self.id_to_pickle_dict_pickle_file=dir_name+"read_ids_per_pickle.p"

        self.split_bool=False
        self.split_fastqs_present=False
        #self.full_id_set_pickle=dir_name+"read_ids_per_pickle.p"

    def check_exists(self):
        
        for p in self.pickle_file_list:
            path_v=Path(p)
            b=os.path.exists(path_v)
            if (b==False):
                return b
        return True

        
    def check_set(self):
        return set(make_dict_from_fasta(self.original_fastq))
    def fastq_to_dict(self,fpath):
        f = open(fpath)
        lines = f.readlines()
        f.close()
        reads={}
        read_lines=[]
        for x in range(len(lines)):

            l=lines[x]
            read_lines.append(l)


            if x%4==0:
                spl=l.split()
                identifier_pre=spl[0]#5/6
                values=identifier_pre.split(":")
                identifier=(values[4]+":"+values[5]+":"+values[6])

            if x%4==3:
                reads[identifier]=read_lines
                read_lines=[]
        return reads

    def split_fastq(self,delete_text_fastqs:bool,temp_delete_idSets_to_pickle_dict:bool):
        print("splitting: ",self.original_fastq)
        big_set=set()

        if self.check_exists():
            
            print("already exists")
            self.split_bool=True
            id_dict=pickle.load(open(self.id_to_pickle_dict_pickle_file,"rb"))

            for k in id_dict.keys():
                #print(k," : ",type(id_dict[k]))

                big_set=big_set.union(id_dict[k])
                #print(len(big_set))
            #print("big set: ",len(big_set))
            return big_set

        fastqsplitter.split_fastqs(self.original_fastq,self.split_list)
        

        for i in range(len(self.split_list)):
            #print(type(self.split_list[i]))
            temp_dict=self.fastq_to_dict(self.split_list[i])
            big_set=big_set.union(set(temp_dict))
            self.id_to_pickle_dict[i]=set(temp_dict.keys())
            pickle.dump(temp_dict,open(self.pickle_file_list[i],"wb"))

        if(delete_text_fastqs):

            for fq in self.split_list:
                os.remove(fq)
            self.split_fastqs_present=False

        else:

            self.split_fastqs_present=True
        self.split_bool=True
        
        """
        this set could be pretty big, could be worth temp_deleting
        """
        
        pickle.dump(self.id_to_pickle_dict,open(self.id_to_pickle_dict_pickle_file,"wb"))
        #self.id_to_pickle_dict.clear()

        #print("in split: ",big_set)
        return big_set
            
    
    """
    takes a dict of reads 
    this method is for efficently organizing the reads based on which pickle
    they are in so pickles only need to be opened once

    union dict a diction representing the sequence to its 

    they will write to the "directory_to_write_to" which is just a string 
    names file as construct name (key in union dict) +"_R"+fastq_id+".fastq

    also clears the union dict to save space

    """
    def super_write_fastqs(self,union_dict:dict,directoy_to_write_to:str,fastq_id:int,sequence_objects:dict,sample_name:str):

        """
        organizes reads into sets per k,
        based on which pickle the read is in 
        """
        sample_fq_dir=directoy_to_write_to+sample_name+"/"
        os.makedirs(sample_fq_dir,exist_ok=True)

        organized_joint_dict={}
        for k in union_dict.keys():
            #print(k)
            joint_set=union_dict[k]

            temp_dictionary_pickle_loc={}

            
            self.id_to_pickle_dict=pickle.load(open(self.id_to_pickle_dict_pickle_file,"rb"))
            #print(self.id_to_pickle_dict.keys())

            for i in range(len(self.pickle_file_list)):
                temp_dictionary_pickle_loc[i]=set()

            
            for ident in joint_set:

                for num in self.id_to_pickle_dict.keys(): 

                    if(ident in self.id_to_pickle_dict[num]):

                        temp_dictionary_pickle_loc[num].add(ident)

            organized_joint_dict[k]=temp_dictionary_pickle_loc
        
        #union_dict.clear()
        """
        now that the reads are organized we have to open each pickle file write fastqs 
        """
        #fqs=[]
        for k in sequence_objects.keys():
            
            if(len(union_dict[k])>1):
                fq=sample_fq_dir+k+"_R"+str(fastq_id)+".fastq"
                fq_buff=open(fq,"wt")
                fq_buff.close()

        for x in range(len(self.pickle_file_list)):

            p_file_object=open(self.pickle_file_list[x],"rb")
            pickled_fastq_dict=pickle.load(p_file_object)
            p_file_object.close()
            
            for k in organized_joint_dict.keys():

                if(len(union_dict[k])>1):

                    #print("writing filtered fastqs"+str(k))
                    fq=sample_fq_dir+k+"_R"+str(fastq_id)+".fastq"
                    #fqs.append(fq)

                    sequence_objects[k].fastqs[fastq_id]=fq

                    specific_reads= organized_joint_dict[k][x]

                    fq_buff=open(fq,"a")

                    for r in specific_reads:
                        lines=pickled_fastq_dict[r]
                        fq_buff.writelines(lines)
                    fq_buff.close()
        #fqs.sort()
    """
    this methods removes the pickles that contain split fastq information
    """
    def destroy_temp_data(self):
        for x in range(len(self.pickle_file_list)):
            os.remove(self.pickle_file_list[x])



        





"""
tolerence is added to both ends of search range

if tolerence==-1 then disregards indexes 
    this will increase computation time
no index ranges and high mismatch threshhold will result in extremely high comp time 
"""

def run_seqkit_grep_function(pattern:str,
                            search_start_ind:int,
                            search_end_index:int,
                            fastq_to_search:str,
                            fastq_to_write:str,
                            threads:int=20,
                            mismatch_threshhold:int=0,
                            append_bool:bool=False,
                            tolerance:int=0,
                            delete_fq:bool=False
                            ):
    """
    1 indexed? 
    
    """
    append_char=">>" if append_bool else ">"

    if(search_start_ind -tolerance<0):
        search_start_ind =0
    else:
        search_start_ind-=tolerance
    search_end_index+=tolerance
    if(search_start_ind<1):
        search_end_index+=abs(search_start_ind)+1
        search_start_ind=1
    cmd=f'seqkit --threads {threads} grep -s -R {search_start_ind}:{search_end_index+tolerance} -p "{pattern}" -m {mismatch_threshhold} -P {fastq_to_search} > {fastq_to_write}'
    print(cmd)#debug-
    return_code=os.system(cmd)#debug
    if (delete_fq):
        os.remove(fastq_to_write)
    return (set(makes_dict_from_fastq(fastq_to_write).keys()))








        
"""
simple fastq dict for retrieveing sequences from fastq
"""


def make_dict_from_fasta(fasta_path) -> dict:
    fa=open(fasta_path,"rt").readlines()
    temp_dict={}

    for i in range(0,len(fa),2):
        temp_dict[fa[i][1:].strip()]=fa[i+1].strip()
    
    return temp_dict
"""
input csv, library that represents each sequence to be dumultiplexed with many different coloumns 

workspace directory that demultiplexing is being done in 

fq1/fq2 path for big fastq which will be demultiplexed? maybe should be removed


this whole method could be replaced with a dataframe that organizes all of these attributes


"""
def make_sequence_objects_from_csv(input_csv,barcode_start,barcode_length,fasta,fastq1_path,fastq2_path,paired,workspace) -> dict:
    
    sequence_object_dict={}
    fasta_dict=make_dict_from_fasta(fasta)
    if (input_csv==""):

        for name in fasta_dict.keys():

            seq=fasta_dict[name]
            rev_seq=reverse_compliment(seq)
            bc=seq[barcode_start :barcode_start+barcode_length]

            rev_barcode=reverse_compliment(bc)
            rev_bc_start=rev_seq.index(rev_barcode)
            rev_bc_end=rev_bc_start+len(rev_barcode)
            

            sequence_object_dict[name]=Sequence_Obj(
                    sequence=fasta_dict[name],
                    fastq1_path=fastq1_path,
                    fastq2_path=fastq2_path,
                    name=name,
                    paired=paired,
                    barcode_start=barcode_start,
                    barcode_end=barcode_start+barcode_length,
                    barcode=bc,
                    rev_barcode=rev_barcode,
                    rev_barcode_start=rev_bc_start,
                    rev_barcode_end=rev_bc_end,
                    secondary_signature_start=-1,
                    secondary_signature_end=-1,
                    secondary_signature=-1,
                    rev_secondary_signature_start=-1,
                    rev_secondary_signature_end=-1,
                    rev_secondary_signature=-1,
                    workspace=workspace+name+"/"
                )

    else:
        df=pd.read_csv(input_csv)

        
        #fasta_dict=make_dict_from_fasta(fasta)

        #sequence_object_dict={}
        cols=set(df.columns)

        if(barcode_start==barcode_length) and ("barcode_start" not in cols):
            raise Exception("no barcode info given")
        
            
        for x in df.index:

            name=df.at[x,"reference"]
            seq=fasta_dict[name]
            rev_seq=reverse_compliment(seq)

            
            if(barcode_start==barcode_length):
                barcode_start=df.at[x,"barcode_start"]
                barcode_length=df.at[x,"barcode_length"]
                bc=seq[barcode_start :barcode_start+barcode_length]
            else:
                bc=seq[barcode_start:barcode_start+barcode_length]
            rev_barcode=reverse_compliment(bc)
            rev_bc_start=rev_seq.index(rev_barcode)
            rev_bc_end=rev_bc_start+len(rev_barcode)
            print("\nhere"+str(cols))
            if("secondary_signature_start" in cols ):
                secondary_sign_start=df.at[x,"secondary_signature_start"]
                secondary_sign_end=secondary_sign_start+df.at[x,"secondary_signature_length"]
                secondary_sign=fasta_dict[name][secondary_sign_start:secondary_sign_end]

                rev_sec_sign=reverse_compliment(secondary_sign)
                rev_sec_sign_start = rev_seq.index(rev_sec_sign)
                rev_sec_sign_end= rev_sec_sign_start +len(rev_sec_sign)


            else:
                secondary_sign_start = -1
                secondary_sign_end = -1
                secondary_sign = -1

                rev_sec_sign = -1
                rev_sec_sign_start = -1
                rev_sec_sign_end = -1

        

            

            sequence_object_dict[name]=Sequence_Obj(
                sequence=fasta_dict[name],
                fastq1_path=fastq1_path,
                fastq2_path=fastq2_path,
                name=name,
                paired=paired,
                barcode_start=barcode_start,
                barcode_end=barcode_start+len(bc),
                barcode=bc,
                rev_barcode=rev_barcode,
                rev_barcode_start=rev_bc_start,
                rev_barcode_end=rev_bc_end,
                secondary_signature_start=secondary_sign_start,
                secondary_signature_end=secondary_sign_end,
                secondary_signature=secondary_sign,
                rev_secondary_signature_start=rev_sec_sign_start,
                rev_secondary_signature_end=rev_sec_sign_end,
                rev_secondary_signature=rev_sec_sign,
                workspace=workspace+name+"/"
            )
    return sequence_object_dict
"""
seqkit grep is 1 indexed and inclusive of the final value of its range 



clipped int that represents how much 
fastq_id fastq 1 or fastq 2
"""

"""
runs grep and accepts a clipped argument and appends the set to the main dictionary

index tolerence can only apply to the initial but there are cases where that could false 
"""
def run_multi_greps(read_id_dict:dict,clipped:int,index_tolerence:int,delete_fastqs:bool,mismatches_allowed:int,pattern_type:str,pattern:str,pattern_start:int,pattern_end:int,fastq:str,seq_folder:str,front:bool=False):
    #debug#print("running multi_greps for ", pattern_type)
    new_reads_dict={}
    folder=seq_folder
    #file_name_start=
    new_reads_dict["init_"+pattern_type]=run_seqkit_grep_function(pattern=pattern,search_start_ind=pattern_start,search_end_index=pattern_end,fastq_to_search=fastq,fastq_to_write=seq_folder+"init_"+pattern_type+".fastq",mismatch_threshhold=mismatches_allowed,tolerance=index_tolerence)
    #if(len(pattern)-pattern_end)<clipped:
    append_these=[seq_folder+"init_"+pattern_type+".fastq"]
    clipped+=1
    #print("initial grep run")#debug
    if not front:
        for x in range(1,clipped):
            
            append_these.append(seq_folder+pattern_type+"_clipped_"+str(x)+".fastq")
            new_reads_dict[pattern_type+"_clipped_"+str(x)]=run_seqkit_grep_function(pattern=pattern[:len(pattern)-x],search_start_ind=pattern_start+x,search_end_index=pattern_end,fastq_to_search=fastq,fastq_to_write=seq_folder+pattern_type+"_clipped_"+str(x)+".fastq",mismatch_threshhold=0,tolerance=0)
    else:
        for x in range(1,clipped):
            append_these.append(seq_folder+pattern_type+"_clipped_"+str(x)+".fastq")
            new_reads_dict[pattern_type+"_clipped_"+str(x)]=run_seqkit_grep_function(pattern=pattern[x:],search_start_ind=pattern_start,search_end_index=pattern_end-x,fastq_to_search=fastq,fastq_to_write=seq_folder+pattern_type+"_clipped_"+str(x)+".fastq",mismatch_threshhold=0,tolerance=0)
    #deug#print("clipped run")
    #for k in new_reads_dict.keys():
    #keys=list(new_reads_dict.keys())

    for x in range(1,clipped):

        new_reads_dict[pattern_type+"_clipped_"+str(x)]-=new_reads_dict["init_"+pattern_type]
        for i in range(x+1,clipped):
            new_reads_dict[pattern_type+"_clipped_"+str(x)]-=new_reads_dict[pattern_type+"_clipped_"+str(x)]
    #debug#print("organizing reads")
    for k in new_reads_dict.keys():
        read_id_dict[k]=new_reads_dict[k]
    #debug#print("appending reads")
    cmd="cat "
    return_fastq=seq_folder+pattern_type+ "_appended.fastq"
    #print(append_these  )
    for x in range(len(append_these)):
        cmd+=(append_these[x] + " ")
    cmd+="> "+ seq_folder+pattern_type+ "_appended.fastq"
    
    os.system(cmd)
    #debug#print("appended!")
    return return_fastq

    
    
def append_files(files,new_file_name):
    cmd="cat "
    for x in range(len(files)):
        cmd+=(files[x] + " ")
    cmd+="> "+ new_file_name
    os.system(cmd)



def run_seqkit_grep(sequence_object:Sequence_Obj,clipped:int,rev_clipped:int,index_tolerence:int,delete_fastqs:bool,fastq_id:int,mismatches_allowed:int):
    #f=open(str(sequence_object.sample_name)+"_test.txt","wt")
    #f.write(str(vars(sequence_object)))
    fastq=sequence_object.fastq_paths[fastq_id]
    #makes folder for demultiplex
    os.makedirs(sequence_object.workspace,exist_ok=True)

    fastq_folder=sequence_object.workspace+"fq"+str(fastq_id)+"/"
    os.makedirs(fastq_folder,exist_ok=True)

    #fastqs
    
    fastq_unfiltered=fastq_folder + "unfiltered.fastq"


    fastq_sec_filtered=fastq_folder + "sec_fitlered_filtered.fastq"
    fastq_rev_sec_filtered=fastq_folder + "rev_sec_filtered.fastq"
    
    fastq_filtered=fastq_folder + "filtered_out.fastq"

    read_ids={}
    bc_fastq=run_multi_greps(read_id_dict=read_ids,clipped=clipped,index_tolerence=index_tolerence,delete_fastqs=delete_fastqs,mismatches_allowed=mismatches_allowed,pattern_type="barcode",pattern=sequence_object.barcode,pattern_start=sequence_object.barcode_start,pattern_end=sequence_object.barcode_end,fastq=fastq,seq_folder=fastq_folder)
    #debug#print("first_multi_run")
    rev_bc_fastq=run_multi_greps(read_id_dict=read_ids,clipped=0,index_tolerence=index_tolerence,delete_fastqs=delete_fastqs,mismatches_allowed=mismatches_allowed,pattern_type="rev_barcode",pattern=sequence_object.rev_barcode,pattern_start=sequence_object.rev_barcode_start,pattern_end=sequence_object.rev_barcode_end,fastq=fastq,seq_folder=fastq_folder)

    append_files([bc_fastq,rev_bc_fastq],fastq_unfiltered)
    read_ids["unfiltered"]=set(makes_dict_from_fastq(fastq_unfiltered).keys())

    
    complete_set=set()

    if(sequence_object.secondary_signature!=-1):
        v_threshold=0
        v_section=pattern=sequence_object.secondary_signature
        if(len(v_section)>20):
            v_threshold=4
        elif(len(v_section)<=20 and len(v_section)>15 ):
            v_threshold=3
        elif(len(v_section)<=15 and len(v_section)>=5 ):
            v_threshold=2
        elif(len(v_section)<5):
            v_threshold=1
    
        secondary_sign_mismatches=v_threshold#TODO hardcoded :(


        read_ids["sec_filter"]=run_seqkit_grep_function(pattern=sequence_object.secondary_signature,search_start_ind=sequence_object.secondary_signature_start,search_end_index=sequence_object.secondary_signature_end,fastq_to_search=fastq_unfiltered,fastq_to_write=fastq_sec_filtered,mismatch_threshhold=secondary_sign_mismatches,tolerance=2)
        read_ids["rev_sec_filter"]=run_seqkit_grep_function(pattern=sequence_object.rev_secondary_signature,search_start_ind=sequence_object.rev_secondary_signature_start,search_end_index=sequence_object.rev_secondary_signature_end,fastq_to_search=fastq_unfiltered,fastq_to_write=fastq_rev_sec_filtered,mismatch_threshhold=secondary_sign_mismatches,tolerance=2)
        complete_set=read_ids["sec_filter"].union(read_ids["rev_sec_filter"])
        read_ids["reads_lost_to_filter"]=read_ids["unfiltered"]-complete_set
        #complete_set=read_ids["sec_filter"].union(read_ids["rev_sec_filter"])
    else:
        complete_set=read_ids["unfiltered"]
    #here the sets need to be compared in order to identify which ids were unique to which search
    #TODO
    
    #print(vars(sequence_object))
    #complete_set=read_ids["sec_filter"].union(read_ids["rev_sec_filter"])

    complete_set_pickle=fastq_folder+"complete_set_of_reads.p"
    pickle.dump(complete_set,open(complete_set_pickle,"wb"))
    read_id_data_pickle=fastq_folder+"read_id_data.p"
    pickle.dump(read_ids,open(read_id_data_pickle,"wb"))
    #print(read_ids)

    open(sequence_object.workspace+"grepped.txt","wt").write("done")


"""
latest and greatest
"""
def check_done(sequence_folder:str) -> bool:
    #print("looking here: ",sequence_folder+"grepped.txt")

    comp_set_fq1="fq1/complete_set_of_reads.p"

    if  (os.path.exists(sequence_folder+"fq1/complete_set_of_reads.p"))\
    and (os.path.exists(sequence_folder+"fq2/complete_set_of_reads.p")) \
        \
    and (os.path.exists(sequence_folder+"fq1/read_id_data.p"))\
    and (os.path.exists(sequence_folder+"fq2/read_id_data.p")) \
        \
    and (os.path.exists(sequence_folder+"grepped.txt")):

        return True
    else:
        return False

def check_all_done(seq_objects:dict()):
    return_dict={}

    
    for k in seq_objects.keys():
        print("checking: "+k,end="\r")
        if not (check_done(seq_objects[k].workspace)):
            
            print("check done not done! ",check_done(seq_objects[k].workspace))
            return_dict[k]=seq_objects[k]
    return return_dict

def grep_both_fastq(sequence_object:Sequence_Obj,clipped:int,rev_clipped:int,index_tolerence:int,delete_fastqs:bool,mismatches_allowed:int):

    run_seqkit_grep(sequence_object=sequence_object, clipped=clipped, rev_clipped=rev_clipped, index_tolerence=index_tolerence, delete_fastqs=delete_fastqs, mismatches_allowed=mismatches_allowed,fastq_id=1)
    run_seqkit_grep(sequence_object=sequence_object, clipped=clipped, rev_clipped=rev_clipped, index_tolerence=index_tolerence, delete_fastqs=delete_fastqs, mismatches_allowed=mismatches_allowed,fastq_id=2)

def parallel_grepping(sequence_objects:dict,fwd_clips:int,rev_clips:int,index_tolerence:int,delete_fastq:bool,paired:bool=True,mismatches:int=0,threads=10,iteration:int=0,overwrite:bool=True):
    """
    runs grep in parallel 
    """
    itr_val=iteration
    print("iteration value:XXX ",itr_val)
    countx=0
    procs=[]
    seq_keys=list(sequence_objects.keys())
    seq_count=0
    THREADS=threads
    #print(seq_keys)
    run_all=True
    
    while(seq_count<len(seq_keys)+6):
        
        print(seq_count)
        procs=[]
            
        seq_index=seq_count
        print(range(seq_index,seq_index+THREADS))
        for i in range(seq_index,(seq_index+THREADS)):
            #seq_index+i
            #print(seq_index)
            if(i<len(seq_keys)):
                seq_keys[i]
                if(not overwrite):
                    previously_run=check_done(sequence_objects[seq_keys[i]].workspace)
                else:
                    previously_run=False
                #print(previously_run)
                if(previously_run and not overwrite):
                    print("grepped")
                else:
                    #(sequence_object:Sequence_Obj,clipped:int,rev_clipped:int,index_tolerence:int,delete_fastqs:bool,mismatches_allowed:int)
                    x=multiprocessing.Process(target=grep_both_fastq, args=(sequence_objects[seq_keys[i]],fwd_clips,rev_clips,index_tolerence,delete_fastq,mismatches))
                    x.start()
                    procs.append(x)
            #seq_index=+1
            else:
                #info#print("no more to run")
                pass

        for p in procs:
            p.join()


        #time.sleep(1000)
        start=time.time()
        seq_count+=THREADS
        itr_val+=1
    
    print("checking")
    not_done_dict=check_all_done(sequence_objects)
    #print(not_done_list)

    while len(not_done_dict) > 0 and itr_val >4:
        print("rip")
        itr_val=iteration
        #itr_val=iteration
        #not_done_dict=check_all_done(sequence_objects)
        parallel_grepping(sequence_objects=not_done_dict,fwd_clips=fwd_clips,rev_clips=rev_clips,index_tolerence=index_tolerence,delete_fastq=delete_fastq,iteration=itr_val+1,mismatches=mismatches)
    if(itr_val>4):
        print("could not finish some of the reads: ",list(not_done_dict.keys()))
    
        #warning could not finish one of the reads
    return not_done_dict    

def regular_grepping(sequence_objects:dict,fwd_clips:int,rev_clips:int,index_tolerence:int,delete_fastq:bool,paired:bool=True,mismatches:int=0,iteration:int=0,overwrite:bool=False):
    """
    runs grep in parallel 
    """
    print("regular grepping")
    itr_val=iteration

    for k in sequence_objects.keys():
            previously_run=check_done(sequence_objects[k].workspace)
            if(previously_run and (not overwrite)):
                print("grepped")
            else:
                grep_both_fastq(sequence_objects[k],fwd_clips,rev_clips,index_tolerence,delete_fastq,mismatches)



    not_done_dict=check_all_done(sequence_objects)


    while len(not_done_dict) > 0 and itr_val >4:
        itr_val=iteration
        """if itr_val>0:
            not_done_dict=check_all_done(sequence_objects)"""
        
        #not_done_dict=check_all_done(sequence_objects)
        regular_grepping(sequence_objects=not_done_dict,fwd_clips=0,rev_clips=0,index_tolerence=0,delete_fastq=False,iteration=itr_val+1,mismatches=mismatches)
    if(itr_val>4):
        print("could not finish one of the reads: ",list(not_done_dict.keys()))


"""
checks each sequence for a grepped.txt and returns true if found 
"""


def finds_multigrepped_reads(sequence_objects:dict,remove:bool=True,resolve:bool=False,print_multi_grep_dict:bool=True,demultiplex_workspace:str=None) -> dict:
    """
    filters reads based on weather or not they map to multiple constructs 
    returns a dictionary mapping each read to a list of the reads it mapped to
    """
    union_dictionary={}
    fq1_sets={}
    fq2_sets={}

    for k in sequence_objects.keys():


        #print(k)
        seq_object_fastq_folder=sequence_objects[k].workspace
        fq1_complete_set_pickle=seq_object_fastq_folder+"fq1/complete_set_of_reads.p"
        fq2_complete_set_pickle=seq_object_fastq_folder+"fq2/complete_set_of_reads.p"

        fq1_complete_set=pickle.load(open(fq1_complete_set_pickle,"rb"))
        fq2_complete_set=pickle.load(open(fq2_complete_set_pickle,"rb"))

        fq1_sets[k]=fq1_complete_set
        fq2_sets[k]=fq2_complete_set
        #joint set of ids from complete sets of fqs
        union_dictionary[k]=fq1_complete_set.union(fq2_complete_set)
        """
        now that all reads are collected into a big dictionary of sets 
        we can see which ones are mapped to multiple reads
        """
    used_reads={}

    for k in sequence_objects.keys():

        for read in union_dictionary[k]:

            if read not in used_reads.keys():
                used_reads[read]=[k]
            else:
                used_reads[read].append(k)
    
    multi_grep_dict={}

    
    for read in used_reads.keys():
        used_by_list=used_reads[read]


        if len(used_by_list)>1:
            temp_list=[]
            #multi_grep_dict[read]=used_by_list
            
            for name in used_by_list:

                if(read in fq1_sets[name] and read in fq2_sets[name]):
                    temp_list.append([0,0,1])
                elif(read in fq1_sets[name] and read not in fq2_sets[name]):
                    temp_list.append([1,0,0])
                elif(read not in fq1_sets[name] and read in fq2_sets[name]):
                    temp_list.append([0,1,0])
                else:
                    temp_list.append([0,0,0])

                if (remove):
                    union_dictionary[name].remove(read)
            multi_grep_dict[read]=[used_by_list,temp_list]

    if(print_multi_grep_dict):
        pickle.dump(multi_grep_dict,open(demultiplex_workspace+"multigrepped_reads.p","wb"))

    return union_dictionary

def resolve_or_analyze_multigrepped_reads(union_sets:dict,remove:bool=True,resolve:bool=False):
    pass
 

def create_report(sequence_objects:dict,fq1:str,fq2:str,working_directory:str,unioned_sets:dict):


    orginal_len={}
    mixed_total_dict={}

    fastqs={}
    dict_of_lists={}
    FQS=["fq1","fq2"]
    #basic_info=["construct_name_","final_count_","lost_to_multigrep_",]
    temp_read_data_pickle=sequence_objects[list(sequence_objects.keys())[0]].workspace+"fq1/read_id_data.p"
    read_data_keys=list(pickle.load(open(temp_read_data_pickle,"rb")).keys())


    for fq in FQS:
        if(fq=="fq1"):
            fastqs[fq]=fq1
        else:
            fastqs[fq]=fq2

        mixed_total_dict[fq]=set(makes_dict_from_fastq(fastqs[fq]))
        orginal_len[fq]=len(mixed_total_dict[fq])
        print("orginal len: ",orginal_len[fq])
        dict_of_lists[fq+"_construct_name_"]=[]
        dict_of_lists[fq+"_final_count_"]=[]
        dict_of_lists[fq+"_lost_to_multigrep_"]=[]
        for data_key in read_data_keys:
            dict_of_lists[fq+"_"+data_key]=[]


        

    for k in sequence_objects.keys():
        #sequence_objects
        seq_obj=sequence_objects[k]
        final_reads=unioned_sets[k]

        print("final_counts: ",len(final_reads))

        for fq in FQS:
            mixed_total_dict[fq]-=final_reads
            fq_complete_set=pickle.load(open(seq_obj.workspace+fq+"/complete_set_of_reads.p","rb"))
            fq_read_data=pickle.load(open(seq_obj.workspace+fq+"/read_id_data.p","rb"))
            
            dict_of_lists[fq+"_construct_name_"].append(k)
            dict_of_lists[fq+"_final_count_"].append(len(final_reads))
            dict_of_lists[fq+"_lost_to_multigrep_"].append(len(fq_complete_set)-len(final_reads))

            for data_key in read_data_keys:
                dict_of_lists[fq+"_"+data_key].append(len(fq_read_data[data_key]))
    df=pd.DataFrame()

    for k in dict_of_lists.keys():
        df[k]=dict_of_lists[k]
    print("len: mixed: ",len(mixed_total_dict[FQS[0]]))
    print("len: mixed: ",len(mixed_total_dict[FQS[0]])," / ",orginal_len[FQS[0]]," = ",len(mixed_total_dict[FQS[0]])/orginal_len[FQS[0]])
    len_col=[" "]*len(sequence_objects.keys())
    len_col[0]=len(mixed_total_dict[FQS[0]])
    df["percent_reads_used"]=[(1-(len(mixed_total_dict[FQS[0]])/orginal_len[FQS[0]]))*100]*len(sequence_objects.keys())
    df["total_read_count"]=len_col

    print(working_directory+"demultiplex_info.csv")
    df.to_csv(working_directory+"demultiplex_info.csv",index=False)



"""
split is default to 10. disregarding extremes, the higher the split the lighter the memeory load
library csv 
    each construct must have a secondary signiture start index and len in order to process, 
    barcode given in main arguements 
"""
def demultiplex_run(library_csv,demulti_workspace,report_folder,mixed_fastq1,mixed_fastq2,fasta,barcode_start=0,barcode_length=0,split:int=10,clipped:int=0,rev_clipped:int=0,index_tolerance:int=0,parallel:bool=False,mismatch_tolerence:int=0,overwrite:bool=False):

    if(type(mixed_fastq1)==type(("x","x"))):
        mixed_fastq1=mixed_fastq1[0]
        mixed_fastq2=mixed_fastq2[0]
    sample_name=mixed_fastq1.split("_R1")[0].split("/")[-1]

    """
    makes dictionary of sequence objects
    """
    #print("demulti_workspace: ",demulti_workspace)
    temp_ws=demulti_workspace+"/"+sample_name+"_demultiplex_folders_and_files/"
    #final_sample_folder=temp_ws+"sample_fqs/"

    #print(temp_ws)

    print(repr(temp_ws))
    os.makedirs(temp_ws,exist_ok=True)
    #print("tempworkspace: ",temp_ws)
    """
    all the little stuff gets stored per sequence here, at least temporarily 
    """
    seq_data_folder=temp_ws+"sequence_data/"
    
    os.makedirs(seq_data_folder,exist_ok=True)

    sequence_objects=make_sequence_objects_from_csv(
        input_csv=library_csv,
        barcode_start=barcode_start,
        barcode_length=barcode_length,
        fasta=fasta,
        fastq1_path=mixed_fastq1,
        fastq2_path=mixed_fastq2,
        paired=True,
        workspace=seq_data_folder)

    #print("\nhere:\n"+(sequence_objects))
    #print("workspace: ",vars(sequence_objects["3042-O-flank_1=hp1-DB"]))
    #demultiplex_workspace=demulti_workspace#"demultiplexed_sequences/"
    """
    makes a super fastq for memory efficent access to fastq reads

    checks values of ids to verify that the two fastq provided contain matching paired end reads
    """  
    super_fq1=super_fastq(mixed_fastq1,split,super_dir=temp_ws)
    super_set1=super_fq1.split_fastq(True,False)

    super_fq2=super_fastq(mixed_fastq2,split,super_dir=temp_ws)
    super_set2=super_fq2.split_fastq(True,False)
    #print(super_set2)
    c=len(super_set1.difference(super_set2))

    super_set1.clear()
    super_set2.clear()

    if c!=0:
        raise Exception("Fastq ids do not match, please verify that the file")


    

    """
    runs grep in parallel 
    """
    if(parallel):
        parallel_grepping(sequence_objects=sequence_objects,fwd_clips=clipped,rev_clips=rev_clipped,index_tolerence=index_tolerance,delete_fastq=False,mismatches=mismatch_tolerence,overwrite=overwrite)
    else:
        regular_grepping(sequence_objects=sequence_objects,fwd_clips=clipped,rev_clips=rev_clipped,index_tolerence=index_tolerance,delete_fastq=False,mismatches=mismatch_tolerence,overwrite=overwrite)
    #print(cum+plus)



    """
    filters reads based on weather or not they map to multiple constructs 
    default is to delete multigrepped reads

    """
    unioned_sets_dictionary=finds_multigrepped_reads(sequence_objects=sequence_objects,demultiplex_workspace=seq_data_folder)
    #debug#print(unioned_sets_dictionary.keys())

    #multigrepped_read_analysis()

    """
    organizes reads for writing
    uses super_fastq object to write fastqs in efficent way
    """

    fq1_paths=super_fq1.super_write_fastqs(unioned_sets_dictionary,report_folder,1,sequence_objects,sample_name=sample_name)
    #super_fq1.destroy_temp_data()

    fq2_paths=super_fq2.super_write_fastqs(unioned_sets_dictionary,report_folder,2,sequence_objects,sample_name=sample_name)
    #super_fq2.destroy_temp_data()

    """
    makes report on what amount of reads were found in which stage
    """
    #sequence_objects:dict,fq1:str,fq2:str,working_directory:str,unioned_sets:dict)
    print("creating report!!!")
    create_report(sequence_objects,mixed_fastq1,mixed_fastq2,report_folder,unioned_sets_dictionary)


    return (),(),(report_folder+sample_name+"/",)



import os.path

import numpy as np
import pandas as pd

from ..util.seq import (EVERY, MATCH, DELET, INS_5, INS_3,
                        SUB_A, SUB_C, SUB_G, SUB_T, SUB_N)

from ..vector.profile import VectorReader


def generate_mut_profile_from_bit_vector(vectors: VectorReader,
                                         clustering_file: str = "",
                                         verbose=False):
    """
    Generate a mutation profile from mutation vectors.

    Parameters
    ----------
    vectors: VectorReader
        Vector reader object

    Returns
    -------
    df_row: dict
        - info: the count per residue of bases for which we have information.
        - cov: the count per residue of bases for which we have coverage.
        - sub_N: the count per residue of bases for which we have mutations.
        - sub_A: the count per residue of bases who are modified to A.
        - sub_C: the count per residue of bases who are modified to C.
        - sub_G: the count per residue of bases who are modified to G.
        - sub_T: the count per residue of bases who are modified to T.
        - sub_rate: the Mutation fraction per residue.
        - num_aligned: the number of aligned reads.
    
    """
    # Convert to a mutation profile
    out = dict()
    out['sequence'] = vectors.seq.decode()
    out['num_aligned'] = vectors.n_vectors
    out["match"] = vectors.count_muts_by_pos(MATCH | INS_5, subsets=True)
    out["sub_A"] = vectors.count_muts_by_pos(SUB_A)
    out["sub_C"] = vectors.count_muts_by_pos(SUB_C)
    out["sub_G"] = vectors.count_muts_by_pos(SUB_G)
    out["sub_T"] = vectors.count_muts_by_pos(SUB_T)
    out["sub_N"] = vectors.count_muts_by_pos(SUB_N, subsets=True)
    out["del"] = vectors.count_muts_by_pos(DELET)
    out["ins"] = vectors.count_muts_by_pos(INS_3, supersets=True)
    # Count non-blank bytes
    out["cov"] = vectors.count_muts_by_pos(EVERY, subsets=True)
    # Unambiguously matching or substituted (informative)
    out["info"] = out["match"] + out["sub_N"]
    # Mutation fraction (fraction mutated among all unambiguously matching/mutated)
    out["sub_rate"] = out["sub_N"] / out["info"]

    muts_per_vector = vectors.count_muts_by_vec(SUB_N, subsets=True)
    hist_margin = 0.5
    hist_min = 0
    hist_max = muts_per_vector.max()
    if np.isnan(hist_max):
        hist_max = 0
    hist_bins = hist_max - hist_min + 1
    out['sub_hist'] = np.histogram(muts_per_vector,
                                   bins=np.linspace(hist_min - hist_margin,
                                                    hist_max + hist_margin,
                                                    hist_bins + 1))[0]

    out['min_cov'] = min(out['cov'])


    # Add clusters
    if clustering_file:
        ext = os.path.splitext(clustering_file)[1]
        if ext == ".json":
            membership = pd.read_json(clustering_file, orient="columns")
        elif ext == ".csv":
            membership = pd.read_csv(clustering_file)
        else:
            raise ValueError(clustering_file)
        cluster_mus = vectors.get_cluster_mus(membership, SUB_N, subsets=True)
        for cluster in cluster_mus.columns:
            out[f"cluster_{cluster}"] = cluster_mus[cluster]

    out.pop("match")
    out.pop('sequence')
    for k in out:
        if isinstance(out[k], pd.Series):
            out[k] = list(out[k])
    return out



from __future__ import annotations
from functools import cached_property, wraps
from io import BufferedReader
from logging import getLogger
from pathlib import Path
from typing import Callable

from ..align.xams import SAM_DELIMITER, SAM_HEADER, sort_xam, view_xam
from ..util import path

logger = getLogger(__name__)


def _reset_seek(func: Callable):
    @wraps(func)
    def wrapper(reader: SamReader, *args, **kwargs):
        prev_pos = reader.sam_file.tell()
        result = func(reader, *args, **kwargs)
        reader.sam_file.seek(prev_pos)
        return result

    return wrapper


def _range_of_records(get_records_func: Callable):
    @wraps(get_records_func)
    @_reset_seek
    def wrapper(reader: SamReader, start: int, stop: int):
        logger.debug(f"Reading records from {reader}, starting at {start} and "
                     f"stopping at {stop}")
        reader.sam_file.seek(start)
        records = get_records_func(reader)
        n_records = 0
        while True:
            if reader.sam_file.tell() < stop:
                n_records += 1
                yield next(records)
            else:
                if reader.sam_file.tell() > stop:
                    raise ValueError(f"SAM reader {reader} stopped at "
                                     f"{reader.sam_file.tell()} but "
                                     f"expected {stop}")
                break
        logger.debug(f"Read {n_records} records from {reader}, starting at "
                     f"{start} and stopping at {stop}")
    return wrapper


@_range_of_records
def _iter_records_single(reader: SamReader):
    """ Yield the read name and line for every read in the file. """
    while line := reader.sam_file.readline():
        yield line.split(SAM_DELIMITER, 1)[0], line, b""


@_range_of_records
def _iter_records_paired_lenient(reader: SamReader):
    prev_line: bytes = b""
    prev_name: bytes = b""
    while line := reader.sam_file.readline():
        if prev_line:
            # Read name is the first field of the line.
            name = line.split(SAM_DELIMITER, 1)[0]
            # The previous read has not yet been yielded.
            if prev_name == name:
                # The current read is the mate of the previous read.
                yield prev_name, prev_line, line
                prev_line = b""
                prev_name = b""
            else:
                # The previous read is paired, but its mate is not
                # in the SAM file. This situation can happen when
                # only the reads mapping to a specific section are
                # exported to a SAM file and only one mate overlaps
                # that section. In this case, it is valid for only
                # one of the two paired mates to be given. Thus,
                # strict mode (which normally ensures that read 2
                # is given if read 1 is paired) is turned off. The
                # given mate is processed as if a single-end read.
                yield prev_name, prev_line, b""
                # Save the current read so that if its mate is the next
                # read, it will be returned as a pair.
                prev_line = line
                prev_name = name
        else:
            # Save the current read so that if its mate is the next
            # read, it will be returned as a pair.
            prev_line = line
            prev_name = line.split(SAM_DELIMITER, 1)[0]
    if prev_line:
        # In case the last read has not yet been yielded, do so.
        yield prev_name, prev_line, b""


@_range_of_records
def _iter_records_paired_strict(reader: SamReader):
    """ Yield the common name and both lines for every pair of reads in
    the file. """
    while line := reader.sam_file.readline():
        yield line.split(SAM_DELIMITER, 1)[0], line, reader.sam_file.readline()


class SamReader(object):
    """ Interface to a SAM file. Extract reads from a specific section
    of a BAM file into a SAM file, then return a SamRecord object for
    each record in the file. """

    def __init__(self, /,
                 xam_inp: (path.OneRefAlignmentInFilePath |
                           path.SectionAlignmentTempFilePath), *,
                 temp_dir: Path,
                 end5: int,
                 end3: int,
                 isfullref: bool,
                 n_procs: int,
                 save_temp: bool):
        self.xam_inp = xam_inp
        self.temp_dir = temp_dir
        self.end5 = end5
        self.end3 = end3
        self.isfullref = isfullref
        self.n_padd = n_procs - 1
        self.save_temp = save_temp
        self.bam_split: Path | None = None
        self.sam_path: path.SectionAlignmentTempFilePath | None = None
        self.sam_file: BufferedReader | None = None

    def __enter__(self):
        # Convert the BAM file to a temporary SAM file
        if self.is_owner:
            # Output only the reads aligning to the section of interest.
            # This step is unnecessary if the section is the full ref.
            if not self.isfullref:
                self.bam_split = path.SectionAlignmentTempFilePath(
                    top=str(self.temp_dir),
                    module=path.Module.VECTOR,
                    step=path.Step.VECTOR_SELECT,
                    sample=self.xam_inp.sample,
                    ref=self.xam_inp.ref,
                    end5=self.end5,
                    end3=self.end3,
                    ext=path.BAM_EXT).path
                view_xam(self.xam_inp.path, self.bam_split, n_padd=self.n_padd,
                         ref=self.xam_inp.ref, end5=self.end5, end3=self.end3)
            # Sort the alignment records by name.
            # Note: this step is unnecessary for single-end reads, but
            # there is currently no way to check whether the reads are
            # single- or paired-end before this step.
            self.sam_path = path.SectionAlignmentTempFilePath(
                top=str(self.temp_dir),
                module=path.Module.VECTOR,
                step=path.Step.VECTOR_SORT,
                sample=self.xam_inp.sample,
                ref=self.xam_inp.ref,
                end5=self.end5,
                end3=self.end3,
                ext=path.SAM_EXT)
            sort_xam(self.xam_inp if self.bam_split is None else self.bam_split,
                     self.sam_path.path, name=True, n_padd=self.n_padd)
            # Delete the temporary BAM file, if it exists.
            if not self.save_temp and self.bam_split is not None:
                self.bam_split.unlink(missing_ok=True)
                self.bam_split = None
        else:
            # If not the owner of the SAM file, then borrow it pre-made.
            self.sam_path = self.xam_inp
        # Open the sorted, section-specific SAM file for reading.
        self.sam_file = open(self.sam_path.path, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sam_file.close()
        self.sam_file = None
        if self.is_owner and not self.save_temp:
            # Delete the SAM file, if its owner.
            self.sam_path.path.unlink(missing_ok=True)
            self.sam_path = None

    @property
    def is_owner(self):
        return isinstance(self.xam_inp, path.OneRefAlignmentInFilePath)

    def _seek_beginning(self):
        """ Seek to the beginning of the SAM file. """
        self.sam_file.seek(0)

    @cached_property
    @_reset_seek
    def _rec1_pos(self):
        """ Return the position of the first record in the SAM file. """
        self._seek_beginning()
        while (line := self.sam_file.readline()).startswith(SAM_HEADER):
            pass
        return self.sam_file.tell() - len(line)

    def _seek_rec1(self):
        """ Seek to the first record in the SAM file. """
        self.sam_file.seek(self._rec1_pos)

    @cached_property
    @_reset_seek
    def paired(self):
        """ Return whether the reads in the SAM file are paired-end. """
        self._seek_rec1()
        first_line = self.sam_file.readline()
        try:
            flag = first_line.split(SAM_DELIMITER, 2)[1]
            paired = bool(int(flag) & 1)
            logger.debug(f"SAM file {self.sam_path} has "
                         f"{'paired' if paired else 'single'}-ended reads")
            return paired
        except (IndexError, ValueError):
            raise ValueError("Failed to determine pairing from first record: "
                             + first_line.decode()+f" {self.sam_path}")

    @property
    def mates_per_record(self):
        """ Mates per record: 1 if single-end, 2 if paired-end. """
        return self.paired + 1

    def iter_records(self, start: int, stop: int, strict_pairs: bool):
        """ Return an iterator of all records in the SAM file. """
        if self.paired:
            if strict_pairs:
                if self.isfullref:
                    return _iter_records_paired_strict(self, start, stop)
                logger.warning(f"Disabling strict pairs for {self} because "
                               f"it is not the full reference sequence")
            return _iter_records_paired_lenient(self, start, stop)
        return _iter_records_single(self, start, stop)

    @_reset_seek
    def iter_batch_indexes(self, records_per_batch: int):
        """ Yield the start and end positions of every batch in the SAM
        file, where each batch should have about ```records_per_batch```
        records. Assume that for nearly all records in paired-end SAM
        files, both mates are present. In the extreme case that only one
        mate is present for every paired-end record, there can be up to
        ```2 * records_per_batch``` records in a batch. """
        logger.debug(f"Computing batches for {self} (paired={self.paired}), "
                     f"targeting {records_per_batch} records per batch")
        if records_per_batch <= 0:
            raise ValueError("records_per_batch must be a positive integer")
        # Number of lines to skip between batches: the number of records
        # per batch minus one (to account for the one line that is read
        # at the beginning of each batch, which ensures that every batch
        # has at least one line) times the number of mates per record.
        n_skip = (records_per_batch - 1) * (self.paired + 1)
        # Start at the beginning of the first record.
        self._seek_rec1()
        # Store sam_file as a local variable to speed up the iteration.
        sam_file = self.sam_file
        # Yield batches until the SAM file is exhausted. If there are no
        # records in the file, then this loop will exit immediately.
        batch = 0
        while line := sam_file.readline():
            # The start position of the batch is the beginning of the
            # line that was just read. Since the position in the file is
            # now the end of that line, the line's length is subtracted.
            start_of_batch = sam_file.tell() - len(line)
            # Read either the prescribed number of lines or to the end
            # of the file, whichever limit is reached first.
            for _, line in zip(range(n_skip), sam_file, strict=False):
                pass
            if self.paired:
                # Get the end position of the current line.
                end_curr = sam_file.tell()
                # Try to get the name of the read in the current line.
                try:
                    name_curr = line.split(SAM_DELIMITER, 1)[0]
                except IndexError:
                    # Getting the read name failed, so it is impossible
                    # to tell whether this read is the mate of the next
                    # read. End this batch at the current read so that
                    # the next iteration can check whether the next read
                    # and the read after the next read are mates.
                    logger.error(f"Cannot find read name in '{line.decode()}'")
                    logger.debug(f"Batch {batch}: {start_of_batch}-{end_curr}")
                    yield start_of_batch, end_curr
                    continue
                # Try to read the next line.
                try:
                    line = next(sam_file)
                except StopIteration:
                    # The end of the file has been reached, so the end
                    # of the current line is the end of the batch.
                    logger.debug(f"Batch {batch}: {start_of_batch}-{end_curr}")
                    yield start_of_batch, end_curr
                    continue
                # Get the end position of the next line.
                end_next = sam_file.tell()
                # Try to get the name of the read in the next line.
                try:
                    name_next = line.split(SAM_DELIMITER, 1)[0]
                except IndexError:
                    # Getting the read name failed, so it is impossible
                    # to tell whether this next read is the mate of the
                    # current read. End this batch at the next read so
                    # that this batch will have an even number of lines,
                    # which, when in doubt, is best for paired reads.
                    logger.error(f"Cannot find read name in '{line.decode()}'")
                    logger.debug(f"Batch {batch}: {start_of_batch}-{end_next}")
                    yield start_of_batch, end_next
                    continue
                # Determine the end of the batch based on the names of
                # the current and the next reads.
                if name_curr == name_next:
                    # If the read names match, then they are mates and
                    # should be part of the same batch. So end the batch
                    # at the end of the next read.
                    logger.debug(f"Batch {batch}: {start_of_batch}-{end_next}")
                    yield start_of_batch, end_next
                else:
                    # Otherwise, the reads are not mates. If the batch
                    # were to end at the end of the next read, then it
                    # would be impossible for the next read to be in the
                    # same batch as the read after the next read, even
                    # if those two reads were mates. To allow those two
                    # reads to be compared and potentially be placed in
                    # the same batch, the current batch must end after
                    # the current read.
                    logger.debug(f"Batch {batch}: {start_of_batch}-{end_curr}")
                    yield start_of_batch, end_curr
                    # Because each iteration of the loop must start at
                    # the position at which the last iteration ended,
                    # move back to that position in the file.
                    sam_file.seek(end_curr)
            else:
                # If the file does not contain paired-end reads, then
                # the end of the batch is the end of the current read.
                end_curr = sam_file.tell()
                logger.debug(f"Batch {batch}: {start_of_batch}-{end_curr}")
                yield start_of_batch, end_curr
            # Increment the batch number.
            batch += 1

    def list_batch_indexes(self, records_per_batch: int):
        """ Return lists of the start and stop indexes of each batch in
        the SAM file. """
        indexes = SamReader.iter_batch_indexes(self, records_per_batch)
        starts, stops = map(list, zip(*indexes, strict=True))
        return starts, stops

    def __str__(self):
        return str(self.sam_path)



from ..util.seq import *
from ..util.util import *
from ..vector.profile import VectorReader


class BitVector:
    """Container object. Contains the name of the reference, the sequence, the bitvector, the read names and the read count.
    """
    
    def __init__(self, path, signal_thresh, include_gu, min_reads, include_del) -> None:
        preprocessing = self.preprocessing(path, signal_thresh=signal_thresh, include_gu=include_gu, min_reads=min_reads, include_del=include_del)

        self.name = path.split('/')[-1][:-(len('.json'))]
        self.sequence = preprocessing[0]
        self.bv = preprocessing[1]
        self.read_index = preprocessing[2]
        self.read_inverse = preprocessing[3]
        self.read_hist = preprocessing[4]
        self.read_names = preprocessing[5]
        self.report = preprocessing[6]
        self.base_to_keep = preprocessing[7]
        self.publish_preprocessing_report(path=path[:-len('.json')]+'_preprocessing_report.txt')
        
    #TODO optimize this 
    def preprocessing(self, path, signal_thresh = 0.005, include_gu = False, min_reads=1000, include_del=False, min_mut_dist = 4):
        """Preprocess the bitvector.
        
        - Remove the bases G and U
        - Count and remove duplicates
        - Remove the bases with a Mutation fraction below signal_thresh
        - save the read names and the read count
        
        Parameters:
        -----------
        
        path: str
            Path to the directory containing the bitvector.
            
        signal_thresh: float
            Mutation fraction below which the bases are removed.
        
        include_gu: bool
            If True, keep the bases G and U.

        min_reads: int
            Minimum number of reads to keep a base.

        include_del: bool
            If True, keep the bases with a deletion.

        max_mut_close_by: int
            Maximum number of mutations close by to keep a base.

        
        Output:
        -------
        
        bv: numpy array (N x D)
            Preprocessed bitvector.
            
        read_hist: numpy array (N)
            Count of reads per bitvector.
            
        read_names: list of str
            Read names.
            
        report: str
            Report of the preprocessing.
            #TODO define
            
        """
        report = {}

        reader = VectorReader.load(path)
        bv = reader.get_all_vectors(numeric=True)

        report['total_number_of_reads'] = bv.shape[0]
        
        ## PER BASE REMOVALS
        # Remove the non-informative bases types
        sequence = reader.seq.decode()
        report['sequence'] = sequence
        
        # Remove the G and U bases
        temp_n_cols = bv.shape[1]
        bases_to_drop = []
        if not include_gu:
            bases_to_drop = [pos for pos, base in zip(bv.columns,
                                                      reader.seq,
                                                      strict=True)
                             if base != A_INT and base != C_INT]
            bv.drop(columns=bases_to_drop, inplace=True)
        report['removed_G_U'] = len(bases_to_drop)

        ## PER READ REMOVALS
        # Remove the bit vectors with too many mutations
        temp_n_reads = bv.shape[0]
        n_muts_per_read = reader.query_vectors(bv, SUB_N | INDEL, subsets=True).sum(axis=1)
        n_muts_per_read_limit = n_muts_per_read.mean() + 3 * n_muts_per_read.std()
        bv.drop(index=bv.index[n_muts_per_read > n_muts_per_read_limit], inplace=True)
        report['too_many_mutations'] = temp_n_reads - bv.shape[0]
        
        # Remove the low-mutation-rate bases
        match_count = (reader.query_vectors(bv, MATCH).sum(axis=0) +
                       reader.query_vectors(bv, MATCH | INS_5).sum(axis=0))
        mut_query = SUB_N | INDEL if include_del else SUB_N
        mut_count = reader.query_vectors(bv, mut_query, subsets=True).sum(axis=0)
        mut_rate = mut_count / (match_count + mut_count)
        temp_n_cols = bv.shape[1]
        bv.drop(columns=bv.columns[mut_rate < signal_thresh], inplace=True)
        report['too_low_mutation_rate'] = temp_n_cols - bv.shape[1]
        report['min_mutation_rate'] = signal_thresh

        # MAKE BV BINARY
        mut_query = SUB_N | INDEL if include_del else SUB_N
        bv = reader.query_vectors(bv, mut_query, subsets=True)

        # Remove the bit mut_vectors with 'max_mut_close_by' consecutive mutations
        muts_too_close = pd.Series(np.zeros(bv.shape[0], dtype=bool),
                                   index=bv.index)
        for pos5 in bv.columns:
            pos3 = pos5 + min_mut_dist - 1
            muts_too_close |= bv.loc[:, pos5: pos3].sum(axis=1) > 1
        temp_n_reads = bv.shape[0]
        bv.drop(index=muts_too_close[muts_too_close].index, inplace=True)
        report['mutations_close_by'] = temp_n_reads - bv.shape[0]

        # Turn bv into a np array
        bases_to_keep = bv.columns - reader.end5  # convert to 0 indexing
        read_names = bv.index.to_list()
        bv = np.array(bv, dtype = np.uint8)
        
        # What's this #TODO
        report['too_few_informative_bits'] = '#TODO'
        
        # Remove the duplicates and count the reads
        bv, read_idx, read_inverse, read_hist = np.unique(bv, axis = 0, return_index=True, return_inverse=True, return_counts = True)
        report['number_of_unique_reads'] = read_hist.shape[0]
        report['number_of_used_reads'] = np.sum(read_hist)
        report['bases_used'] = bv.shape[1]
        
        
        # Sanity check
        assert report['number_of_used_reads'] >= min_reads, "Too few reads after preprocessing"
        assert len(report['sequence']) == report['bases_used'] + report['too_low_mutation_rate'] + report['removed_G_U']
        assert report['total_number_of_reads'] == report['number_of_used_reads'] + report['too_many_mutations'] + report['mutations_close_by']
        return sequence, bv, read_idx, read_inverse, read_hist, read_names, report, bases_to_keep
    

   
   
    def publish_preprocessing_report(self, path):
        """Publish the report in a text file.

        Parameters:
        -----------

        path: str
            Path to the directory where the report will be saved.
            
        """

        report = "Sequence: "+ self.report['sequence'] \
        + "\nLength of the sequence: " + str(len(self.report['sequence'])) \
        + "\nResidues used: " + str(self.report['bases_used']) \
        + "\nResidues removed because they are Gs and Us: " + str(self.report['removed_G_U']) \
        + "\nResidues removed because they have Mutation fraction below {} : {}".format(self.report['min_mutation_rate'], self.report['too_low_mutation_rate']) \
        + "\nNumber of bit mut_vectors in total: "+ str(self.report['total_number_of_reads'])\
        + "\nNumber of bit mut_vectors used: "+ str(self.report['number_of_used_reads']) \
        + "\nNumber of unique bit mut_vectors used: " + str(self.report['number_of_unique_reads']) \
        + "\nNumber of bit mut_vectors discarded: "+ str(self.report['total_number_of_reads'] - self.report['number_of_used_reads']) \
        + "\nBit mut_vectors removed because of too many mutations: " + str(self.report['too_many_mutations']) \
        + "\nBit mut_vectors removed because of too few informative bits: " + str(self.report['too_few_informative_bits']) \
        + "\nBit mut_vectors removed because of mutations close by: " + str(self.report['mutations_close_by'])
        #+ "\nBit mut_vectors removed because of no info around mutations:  " + str(self.report['no_info_around_mutations']
        
        with open(path, 'w') as f:
            print(report)
            f.write(report)
           
           
    def associate_reads_with_likelihoods(self, likelihood_per_read):
        """Associates the reads with their likelihood, using the attributes read_names and read_hist. 
        Publish the reads in a json file.

        Parameters:
        -----------

        likelihood_per_read: array (N x K)
            Likelihood of each read.

        Output:
        -------

        reads: dict
            Dictionary associating the read name with the likelihood.

        """

        reads = {}
        for name, idx in zip(self.read_names, self.read_inverse, strict=True):
            reads[name] = {}
            for k in range(likelihood_per_read.shape[1]):
                reads[name]['K'+str(k+1)] = likelihood_per_read[idx,k]

        return reads


def mutations_bin_arr(bv):
    return query_muts(np.array(bv, dtype=np.uint8), SUB_N | DELET | INS_5 | INS_3, sum_up=False)


def deletions_bin_array(bv):
    return query_muts(np.array(bv, dtype=np.uint8), DELET, sum_up=False)



import pandas as pd
import numpy as np

from  dreem.draw.util import *

import plotly.graph_objects as go
from plotly.offline import plot, iplot

from ..draw import  util
from itertools import cycle
from typing import Tuple, List
from sklearn import metrics
from sklearn.linear_model import LogisticRegression, LinearRegression
from plotly.subplots import make_subplots
import plotly.express as px
from dms_ci import dms_ci



LIST_COLORS = ['red','green','blue','orange','purple','black','yellow','pink','brown','grey','cyan','magenta']

def mutation_fraction(df, show_ci:bool=True)->dict:
    assert len(df) == 1, "df must have only one row"
    mh = df.iloc[0]
    cmap = {"A": "red", "T": "green", "G": "orange", "C": "blue"}  # Color map
    
    traces, layouts = [], []
    mh['index_selected'] = [i+1 for i in range(len(mh['sequence']))] #TODO[i + 1 for i in mh.index_selected] # index starts at 1
    mh_unrolled = pd.DataFrame({'mut_rate':list(mh.sub_rate), 'base':list(mh.sequence), 'index_reset':list(range(len(mh.index_selected))),'index_selected':mh.index_selected, 'paired':list(mh.structure)})

    for bt in set(mh['sequence']):
        df_loc = mh_unrolled[mh_unrolled['base'] == bt]
        if len(df_loc) == 0:
            continue

        hover_attr = pd.DataFrame({'mut_rate':list(df_loc.mut_rate),
                                        'base':list(df_loc.base), 
                                        'index': df_loc['index_selected'],
                                        'paired':[{'.':False, '(':True,')':True}[s] for s in df_loc.paired]})
        traces.append(go.Bar(
            x= np.array(df_loc['index_reset']),
            y= np.array(df_loc['mut_rate']),
            name=bt,
            marker_color=cmap[bt],
            text = hover_attr,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            ))
        if show_ci:
            traces[-1].update(
                        error_y=dict(
                        type='data',
                        symmetric=False,
                        ))

    

    fig = go.Figure(data=traces)

    fig.update_layout(title=f"{mh['sample']} - {mh['reference']} - {mh['section']} - {mh['cluster']} - {mh['num_aligned']} reads",
                        xaxis=dict(title="Position"),
                        yaxis=dict(title="Mutation fraction", range=[0, 0.1]))
   
    fig.update_yaxes(
            gridcolor='lightgray',
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )
    fig.update_xaxes(
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )

    fig.update_xaxes(
            tickvals=mh_unrolled['index_reset'],
            ticktext=["%s %s" % ({'.':'(U)','(':'(P)',')':'(P)'}[x], str(y)) for (x,y) in zip(mh['structure'],mh['index_selected'])],
            tickangle=90,
            autorange=True
    )
    
    # make the background white 
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'df':mh}


def mutation_fraction_identity(data, show_ci:bool=True)->dict:

    assert len(data) > 0, "The combination of sample, reference and section does not exist in the dataframe"
    assert len(data) == 1, "The combination of sample, reference and section is not unique in the dataframe"
    data = data.iloc[0]
    
    df = pd.DataFrame(index = list(data['sequence']))
    fig = go.Figure()
    color_map={'A':'red','C':'blue','G':'yellow','T':'green'}

    data['err_min'] = [dms_ci(p, data['num_aligned'])[0] for p in data['sub_rate']]
    data['err_max'] = [dms_ci(p, data['num_aligned'])[1] for p in data['sub_rate']]

    for base in ['A','C','G','T']:
        df[base] = np.array(data['sub_'+base])/np.array(data['info'])
        fig.add_trace( go.Bar(x=np.arange(len(data['sequence'])), y=list(df[base]), marker_color=color_map[base], showlegend=True, name=base) )
    
    # add error bars to stacked_bar[-1]
    if show_ci:
        fig.data[-1].error_y=dict(
                type='data',
                array= [data['err_max'][i]-data['sub_rate'][i] for i in range(len(data['sequence']))],
                arrayminus = [data['sub_rate'][i]-data['err_min'][i] for i in range(len(data['sequence']))],
                visible=True,
                symmetric=False,
                thickness=1.5,
                width=2,
                color='black'
        )
        
    fig.update_layout(
        title='Mutation fraction identity - {} - {} - {} - {} - {} reads'.format(data['sample'], data['reference'], data['section'], data['cluster'], data['num_aligned']),
        xaxis_title='Position',
        yaxis_title='Mutation fraction',
        )

    fig.update_xaxes(tickangle=0, 
            tickvals=np.arange(len(df.index)), ticktext=list(df.index), tickfont={'size':8})
    

    fig.update_layout(barmode='stack')
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'df':df}


def mutations_in_barcodes(data):
    
    fig = go.Figure()

    len_barcode = len(data['sequence'].iloc[2])
    for sample in data['sample'].unique():
        hist = np.sum(np.stack(data[data['sample']==sample]['sub_hist'].values), axis=0)
        bin_edges = np.arange(0, len_barcode, 1)
        
        fig.add_trace(
            go.Bar(
                x=bin_edges[:-1],
                y=hist,
                name=sample,
                visible=False,
                hovertemplate='Number of mutations: %{x}<br>Number of reads: %{y}<extra></extra>',
                ))
        
    fig.data[0].visible = True
    
    fig.update_layout(barmode='stack', title='Number of mutations in barcodes - {}'.format(data['sample'].unique()[0]))
    fig.update_layout(
        updatemenus=[
            dict(
                active=0,
                buttons=list([
                    dict(label = sample,
                            method = "update",
                            args = [{"visible": [sample == s for s in data['sample'].unique()]},
                                    {"title": 'Number of mutations in barcodes - {}'.format(sample)}])
                    for sample in data['sample'].unique()
                ]), 
                x = 0.7,
                xanchor = 'left',
                y = 1.1,
                yanchor = 'top'
                )
            ])

    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')
    
    return {'fig': fig, 'data': data[['sample','reference','sub_hist']]}


def deltaG_vs_sub_rate(df:pd.DataFrame, models:List[str]=[],  savefile=None, auto_open=False, use_iplot=True, title=None)->dict:

    df_temp = pd.DataFrame()
    for _, row in df.iterrows():
        df_temp = pd.concat([df_temp, pd.DataFrame({'reference':row.reference, 'index':row.index_selected, 'sub_rate':row.sub_rate, 'num_aligned':row.num_aligned, 'deltaG':row['deltaG'],'base':list(row.sequence), 'paired':[s !='.' for s in row.structure]}, index= [i+1 for i in range(len(row.sequence))])])
    
    assert len(df_temp) > 0, "No data to plot"
    df = df_temp.reset_index()

    hover_attr = ['num_aligned','sub_rate','base','index','reference','deltaG']
    tra = {}
    for is_paired, prefix in zip([True,False], ['Paired ','Unpaired ']):
        markers = cycle(list(range(153)))
        x=np.array(df[df.paired == is_paired]['deltaG'])
        y=np.array(df[df.paired == is_paired]['sub_rate'])
        tra[prefix] = go.Scatter(
            x=x,
            y=y,
            text = df[df.paired == is_paired][hover_attr],
            marker_size= df[df.paired == is_paired]['num_aligned']/200,
            mode='markers',
            name=prefix,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            line=dict(color='green' if is_paired else 'red'))
            
        
        for m in models:
            if len(y) > 0:
                fit = util.Fit()
                x_sorted, pred_y_sorted = fit.predict(x,y,m, prefix)
                tra[fit.get_legend()] = go.Scatter(
                    x=x_sorted,
                    y=pred_y_sorted,
                    mode='lines+markers',
                    name=fit.get_legend(), 
                    marker=dict(symbol=next(markers)),
                    line=dict(color='darkseagreen' if is_paired else 'crimson', dash='dash'))

    layout = dict(title = 'Mutation rates of paired / unpaired residues vs the expected energy of the molecule',
            xaxis= dict(title= 'DeltaG',ticklen= 5,zeroline= False),
            yaxis= dict(title= 'Mutation fraction ',ticklen= 5,zeroline= False),
            )

    fig = go.Figure(data=list(tra.values()), layout=layout)
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'df':df}

    
def experimental_variable_across_samples(data:pd.DataFrame, experimental_variable:str, models:List[str]=[])->dict:

    fig = go.Figure()
    
    assert len(data) > 0, "No data to plot"
    assert experimental_variable in data.columns, "Experimental variable not found in data"
    assert len(data['sequence'].unique()) == 1, "More than one sequence found in data. CHeck that reference and section are unique"

    df = pd.DataFrame(
        np.hstack([np.vstack(data['sub_rate'].values), np.vstack(data[experimental_variable].values)]),
        index=data[experimental_variable],
        columns=[c + str(idx+1) for c, idx in zip(data['sequence'].iloc[0], data['index_selected'].iloc[0])] + [experimental_variable]
        ).sort_index()
    
    for col in df.columns:
        if col == experimental_variable:
            continue
        fig.add_trace(
            go.Scatter(
                x = df.index,
                y = df[col],
                mode='lines+markers',
                name=col,
                text= experimental_variable,
                hovertemplate = '<b>Experimental variable: %{x}<br>Mutation fraction: %{y}<extra></extra>',
            ),
        )
            
    fig.update_layout(
        title='Mutation rates across experimental variable - {}'.format(experimental_variable),
        xaxis_title=experimental_variable,
        yaxis_title='Mutation fraction',
        )
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'data':df}



def auc(df:pd.DataFrame,  savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    def make_roc_curve(X, y, y_pred, fig, title):
        fpr, tpr, thresholds = metrics.roc_curve(y, y_pred)
        roc_auc = metrics.auc(fpr, tpr)
        fig.add_trace(go.Scatter(x=fpr, y=tpr,
                            mode='lines',
                            name=title,
                            line=dict(width=2)))
        return fig


    fig = go.Figure()
    for row in df.iterrows():
        X = row[1]['sub_rate'].reshape(-1, 1)
        y = np.array([1 if c == '.' else 0 for c in row[1]['structure']]).reshape(-1, 1)
        y_pred = LogisticRegression().fit(X, y.ravel()).predict_proba(X)[:,1]
        fig = make_roc_curve(X, y, y_pred, fig, row[1]['unique_id'])


    fig.add_trace(
        go.Scatter(x=[0, 1], y=[0, 1], line=dict(color='black', width=2, dash='dash'), showlegend=False)
    )
    fig.update_layout(
        title='ROC Curve',
        xaxis_title='False Positive Rate',
        yaxis_title='True Positive Rate')

    fig.update_yaxes(scaleanchor="x", scaleratio=1)
    fig.update_xaxes(constrain='domain')
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'df':df}


def mutation_fraction_delta(df, savefile=None, auto_open=False, use_iplot=True, title=None)->dict:
    assert len(df) == 2, "df must have 2 row"
    mp_attr = ['sample', 'reference', 'section', 'cluster']
    df['unique_id'] = df.apply(lambda row: ' - '.join([str(row[attr]) for attr in mp_attr]), axis=1)

    mh = pd.Series(
        {
            'sub_rate': df['sub_rate'].values[0] - df['sub_rate'].values[1],
            'sequence': ''.join([c1 if c1 == c2 else '-' for c1,c2 in zip(df['sequence'].values[0],df['sequence'].values[1])]),
            'title': "{} - {} reads vs {} - {} reads".format(df['unique_id'].values[0], df['num_aligned'].values[0], df['unique_id'].values[1], df['num_aligned'].values[1])
        }
    )
    cmap = {"A": "red", "T": "green", "G": "orange", "C": "blue", '-':'grey'}  # Color map
    
    traces, layouts = [], []
    mh_unrolled = pd.DataFrame({'mut_rate':list(mh.sub_rate), 'base':list(mh.sequence), 'index_reset':list(range(len(mh.sequence)))})

    for bt in set(mh['sequence']):
        df_loc = mh_unrolled[mh_unrolled['base'] == bt]
        if len(df_loc) == 0:
            continue

        hover_attr = pd.DataFrame({'mut_rate':list(df_loc.mut_rate),
                                        'base':list(df_loc.base)})
        traces.append(go.Bar(
            x= np.array(df_loc['index_reset']),
            y= np.array(df_loc['mut_rate']),
            name=bt,
            marker_color=cmap[bt],
            text = hover_attr,
            hovertemplate = ''.join(["<b>"+ha+": %{text["+str(i)+"]}<br>" for i, ha in enumerate(hover_attr)]),
            ))
    

    fig = go.Figure(data=traces, 
                    layout=go.Layout(
                        title=go.layout.Title(text=mh['title']),
                        xaxis=dict(title="Position"),
                        yaxis=dict(title="Mutation fraction", range=[0, 0.1])))

    fig.update_yaxes(
            gridcolor='lightgray',
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )
    fig.update_xaxes(
            linewidth=1,
            linecolor='black',
            mirror=True,
            autorange=True
    )
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {'fig':fig, 'df':mh}

               
def _mutations_per_read_subplot(data):

    hist = np.sum(np.stack(data.values), axis=0)
    if (hist[0]==0 and len(hist)<2):
        return go.Bar( x=[], y=[], showlegend=False, marker_color='indianred')
    else:

        bin_edges = np.arange(0, np.max(np.argwhere(hist != 0 )), 1)
        return go.Bar( x=bin_edges, y=hist, showlegend=False, marker_color='indianred')

def mutations_per_read_per_sample(data):

    unique_samples = data['sample'].unique()
    fig = make_subplots(rows=len(unique_samples), cols=1, vertical_spacing=0.4/len(unique_samples),
                        subplot_titles=['Number of mutations per read - {}'.format(sample) for sample in unique_samples])
    for i_s, sample in enumerate(unique_samples):
        
        fig.add_trace(_mutations_per_read_subplot(data[data['sample']==sample]['sub_hist'].reset_index(drop=True)),
                      row=i_s+1, col=1 )
        fig.update_yaxes(title='Count')
        fig.update_xaxes(dtick=10)

    fig.update_layout(autosize=True, height=len(unique_samples)*500, title='Number of mutation per read across samples')
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {
        'fig':fig,
        'data':data
        }
    
def num_aligned_reads_per_reference_frequency_distribution(data):
    assert len(samples:=data['sample'].unique()) == 1, "data must have 1 sample"
    data = data['num_aligned'].values
    return {
            'fig':go.Figure(
                go.Histogram(
                    x=data, 
                    showlegend=False, 
                    marker_color='indianred',
                    hovertemplate="Number of aligned reads: %{x}<br>Count: %{y}<extra></extra>"
                    ),
                layout=go.Layout(
                    title=go.layout.Title(text='{} - Reads per reference count'.format(samples[0])),
                    xaxis=dict(title="Number of aligned reads"),
                    yaxis=dict(title="Count"),
                    plot_bgcolor='white',
                    paper_bgcolor='white'
                    )          
                ),
            'data':data
            }
    
    
def mutation_per_read_per_reference(data):
    assert len(data) == 1, "data must have 1 row"
    data = data.iloc[0]
    sample, reference = data['sample'], data['reference']
    data = data['sub_hist']
    
    # normalize by the number of reads
    fig = px.bar(x=np.arange(0,len(data)), y=data)

    fig.update_layout(barmode='stack')
    fig.update_layout(title='Number of mutations per read - {} - {}'.format(sample, reference))
    fig.update_yaxes(title='Count')
    fig.update_xaxes(title='Number of mutations per read')
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')

    return {
        'fig':fig,
        'data':data
        }



def base_coverage(data):
    
    assert_only_one_row(data)
    data = data.iloc[0]
    
    fig = go.Figure(
        go.Bar(
            x=np.arange(1, 1+len(data['cov'])),
            y=data['cov'],
            showlegend=False,
            marker_color='indianred',
            ),
        layout=go.Layout(
            title=go.layout.Title(text='Base coverage - {} - {}'.format(data['sample'], data['reference'])),
            xaxis=dict(title="Position"),
            yaxis=dict(title="Count")
            )
        )
    fig.update_layout(plot_bgcolor='white',paper_bgcolor='white')
    
    return {'fig':fig, 'data':data}



def compare_mutation_profiles(data, max_plots = 100, max_axis=None):

    # total number of plots
    totNumPlots =  int((data.shape[0] *  (1+data.shape[0]))/2)
    if totNumPlots > max_plots:
        print('Too many plots: {} rows combined together make {} plots when using these arguments. Filtered dataset is: \n\n{}.'.format(data.shape[0], totNumPlots, data[['sample','reference','section','cluster','sub_rate_x']]))
        return {'fig': go.Figure(), 'data': data}
    
    if data.shape[0] == 0:
        print('No data found for this combination of arguments')
        return {'fig': go.Figure(), 'data': data}    

    if data.shape[0] == 1:
        print('Only one row found for this combination of arguments.')
        return {'fig': go.Figure(), 'data': data}        

    # Assert sequence is the same for all rows
    assert data['sequence'].nunique() == 1, 'Sequence is not the same for all rows. Select a subset of the data that has the same sequence.'

    # sort by unique_id
    data = data.sort_values('unique_id').reset_index(drop=True)

    # plot these
    fig = go.Figure()
    
    makePlotLabel = lambda x, y: '{} vs {}'.format(x, y)
    
    traceTrack = []
    annotationTrack = []
    
    for idx1, row1 in data.iloc[:-1].iterrows():
        for _, row2 in data.iloc[idx1+1:].iterrows():
            x, y = row1['sub_rate'], row2['sub_rate']
            xlabel, ylabel = row1['unique_id'], row2['unique_id']
            plotLabel = makePlotLabel(xlabel, ylabel)
            
            # make the plot squared 
            if max_axis is not None:
                maxValue = max_axis
            else:
                maxValue = max(x.max(), y.max(), 0.14) + 0.01
            
            fig.update_xaxes(range=[0, maxValue], constrain='domain')
            fig.update_yaxes(range=[0, maxValue], constrain='domain')
            
            # plot x vs y then the linear regression line, then the 1:1 line, then the R2 and RMSE values
            fig.add_trace(go.Scatter(x=x, y=y, mode='markers', name='mutation fraction', text=plotLabel, visible=False),)
            traceTrack.append(plotLabel)
            
            # linear regression line
            model = LinearRegression()
            model.fit(x.reshape(-1,1), y)
            slope, intercept, r_value = model.coef_[0], model.intercept_, model.score(x.reshape(-1,1), y)
            fig.add_trace(go.Scatter(x=x, y=slope*x+intercept, mode='lines', name='linear regression: y = {}x + {}'.format(round(slope,4), round(intercept,4)), visible=False))
            traceTrack.append(plotLabel)
            
            # 1:1 line
            fig.add_trace(go.Scatter(x=[0, maxValue], y=[0, maxValue], mode='lines', name='1:1 line', visible=False))
            traceTrack.append(plotLabel)
            
            # R2, RMSE and linear regression line
            annot = 'R2 = {} <br> RMSE = {} <br> Lin Reg: y = {}x + {}'.format(round(r_value**2,4), round(np.sqrt(np.mean((y - (slope*x+intercept))**2)),2), round(slope,4), round(intercept,4))
            fig.add_annotation(visible = False, y=0.13,text=annot, showarrow=False)
            annotationTrack.append(annot)
            
            # make the axis equal
            fig.update_layout(
                xaxis=dict(
                    scaleanchor="y",
                    scaleratio=1,
                ))
            
            # add labels
            fig.update_layout(
                xaxis_title="{} sub rate".format(xlabel),
                yaxis_title="{} sub rate".format(ylabel),
                margin=dict(
                    l=0,
                    r=0,
                    b=50,
                    t=50,
                    pad=4
                ),
                title = plotLabel + ' mutation fraction correlation',
            )
            
    # add button to select a specific plot using the traceTrack and a dropdown menu
    fig.update_layout(
        updatemenus=[
            dict(
                active=0,
                buttons=list([
                    # make the annotation visible and the plot visible
                    dict(label=plot,
                            method="update",
                            args=[{"visible": [True if traceTrack[j] == plot else False for j in range(len(traceTrack))]},
                                    {"annotations": [dict(visible = True if np.unique(traceTrack)[j] == plot else False, y=0.13,text=annotationTrack[idx], showarrow=False) for j in range(len(annotationTrack))]}
                                  ])
                    for idx, plot in enumerate(np.unique(traceTrack))
                ]),
                x = 1,
                y = 0.4,
                xanchor = 'left',
                yanchor = 'top',
            )
        ]
    )
    
    # make the first plot visible
    fig.data[0].visible = True
    fig.data[1].visible = True
    fig.data[2].visible = True

    # make the first annotation visible
    fig.layout.annotations[0].visible = True
    
    return {'fig': fig, 'data': data}


"""Draw DREEM output and draw conclusions.
"""
import dreem.draw.manipulator as manipulator
import dreem.draw.plotter as plotter
import dreem.draw.util as util
from dreem.draw.study import Study as Study
from dreem.draw.load_dataset import load_dataset as load_dataset

from dreem.draw.main import run, cli, params



DEFAULT_DOCSTRING_INDENT = 2

def find_offset(docstring):
    docstring = clean_docstring(docstring)
    min_indent = 100
    for line in docstring.split('\n')[1:]:
        if is_line_empty(line):
            continue
        def count_indent(line):
            for count, char in enumerate(line):
                if char != '\t':
                    return count
        min_indent = min(min_indent, count_indent(line))
    return min_indent

def docstring_to_dict_google(docstring, offset=DEFAULT_DOCSTRING_INDENT):    
    docstring = clean_docstring(docstring)
    
    out = {}
    
    out['header'] = docstring.split('\n')[0]
    
    current_section = None
    for line in docstring.split('\n'):
        if is_line_empty(line):
            continue
        if line.startswith('\t'*(1+offset)) and current_section is not None:
            attr = line.strip()
            out[current_section][line.strip().split(' ')[0]] = ' '.join(line.strip().split(' ')[1:])
        elif line.startswith('\t'*offset):
            current_section = line.strip()
            out[current_section] = {}
    
    return out


def dict_google_to_docstring(d, offset=DEFAULT_DOCSTRING_INDENT):
    
    docstring = ''
    if 'header' in d.keys():
        docstring = d.pop('header')

    docstring += '\n'
    for key in d.keys():
        docstring += '\n' + '\t'*offset + key + '\n'
        for item in d[key]:
            docstring += '\t'*(1+offset) + item + ' ' + d[key][item] + '\n'
    
    return docstring


def style_child_takes_over_parent(prnt_doc, child_doc): 

    docs = [prnt_doc, child_doc]
    for idx, doc in enumerate(docs):
        docs[idx] = docstring_to_dict_google(doc)
    prnt_doc, child_doc = docs
    
    for key in prnt_doc.keys():
        if key not in child_doc.keys():
            child_doc[key] = prnt_doc[key]
        else:
            if key == 'header':
                continue
            for item in prnt_doc[key]:
                if not item in child_doc[key].keys():
                    child_doc[key][item] = prnt_doc[key][item]

    return dict_google_to_docstring(child_doc)

def is_line_empty(line):
    return line.replace('\t', '').replace(' ', '') == ''

def clean_docstring(docstring):
    return docstring.replace( '    ', '\t')


import os, sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
import os
import pytest
from dreem.main import run
from dreem.util.files_sanity import compare_fields
import json
import shutil

dreem_root = os.path.dirname(os.path.dirname(__file__))
test_files = os.path.join(dreem_root, 'test_files')
top_dir = test_files.replace('test_files', 'test_output')
temp_dir = test_files.replace('test_files', 'test_temp')
samples = ['my_python_sample', 'my_cli_sample']

FIELDS_FROM_ALIGNMENT_REPORT = ['num_reads', 'skips_low_mapq', 'skips_short_reads', 'skips_too_many_muts']

def get_expected_file(sample):
    return json.load(open(os.path.join(test_files, sample, '{}.json'.format(sample)), 'r'))

def get_output_file_path(sample):
    return os.path.join(top_dir, '{}.json'.format(sample))

def output_file(sample):
    assert os.path.exists(get_output_file_path(sample))
    return json.load(open(get_output_file_path(sample), 'r'))


def references_from_json(json_file):
    return [k for k in json_file.keys() if type(json_file[k]) == dict]


def attribute_from_reference(json_file, ref):
    return [k for k in json_file[ref].keys() if type(json_file[ref][k]) != dict]


def section_from_reference(json_file, ref):
    return [k for k in json_file[ref].keys() if type(json_file[ref][k]) == dict]


def attribute_from_section(json_file):
    reference = references_from_json(json_file)[0]
    section = section_from_reference(json_file, reference)[0]
    return [k for k in json_file[reference][section].keys() if type(json_file[reference][section][k]) != dict]


def attribute_from_pop_avg(json_file):
    reference = references_from_json(json_file)[0]
    section = section_from_reference(json_file, reference)[0]
    return json_file[reference][section]['pop_avg'].keys()


def test_clear_output():
    if os.path.exists(top_dir):
        shutil.rmtree(top_dir)
    os.makedirs(top_dir)
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.makedirs(temp_dir)

@pytest.mark.parametrize('sample', ['my_python_sample'])
def test_run_python(sample):
    run(
        out_dir=top_dir,
        temp_dir=temp_dir,
        fastq1=(os.path.join(test_files, sample, '{}_R1.fastq'.format(sample)),),
        fastq2=(os.path.join(test_files, sample, '{}_R2.fastq'.format(sample)),),
        fasta=os.path.join(test_files, sample, '{}.fasta'.format(sample)),
        library=os.path.join(test_files, sample, 'library.csv'),
        samples=os.path.join(test_files, sample, 'samples.csv'),
        rerun=True,
        flat=True,
        #rnastructure_path='/Users/ymdt/src/RNAstructure/exe/',
    )

@pytest.mark.parametrize('sample', ['my_cli_sample'])
def test_run_cli(sample):
    os.system('dreem --out-dir {} --temp-dir {} --fastq1 {} --fastq2 {} --fasta {} --library {} --samples {} --section roi --section ms2 --section 1-50 --rerun '.format(
        top_dir,
        temp_dir,
        os.path.join(test_files, sample, '{}_R1.fastq'.format(sample)),
        os.path.join(test_files, sample, '{}_R2.fastq'.format(sample)),
        os.path.join(test_files, sample, '{}.fasta'.format(sample)),
        os.path.join(test_files, sample, 'library.csv'),
        os.path.join(test_files, sample, 'samples.csv'),
    ))

@pytest.mark.parametrize('sample', samples)
def test_output_exists(sample):
    assert os.path.exists(os.path.join(os.getcwd(), 'test_output', sample+'.json'))


@pytest.mark.parametrize('sample,attr', [(s,k) for s in samples for k in get_expected_file(s).keys() if type(get_expected_file(s)[k]) != dict])
def test_sample_attributes(sample, attr):
    if not type(get_expected_file(sample)[attr]) == dict:
        compare_fields(get_expected_file(sample), output_file(sample), [attr])


@pytest.mark.parametrize('sample,reference,attr', [(s, c, a) for s in samples \
                                                             for c in list(references_from_json(get_expected_file(s))) 
                                                             for a in attribute_from_reference(get_expected_file(s), c)])
def test_library_attributes(sample, reference, attr):
    if attr in FIELDS_FROM_ALIGNMENT_REPORT:
        pytest.skip('skipped')
    else:
        compare_fields(get_expected_file(sample), output_file(sample), [reference, attr])


@pytest.mark.parametrize('sample,reference,section,attr', [(ss, c, s, a) for ss in samples \
                                                                 for c in list(references_from_json(get_expected_file(ss))) 
                                                                 for s in section_from_reference(get_expected_file(ss), c)
                                                                 for a in attribute_from_section(get_expected_file(ss))])
def test_sections_attributes(sample, reference, section, attr):
    if attr in FIELDS_FROM_ALIGNMENT_REPORT:
        pytest.skip('skipped')
    else:
        compare_fields(get_expected_file(sample), output_file(sample), [reference, section, attr])


@pytest.mark.parametrize('sample,reference,attr', [(s, c, a) for s in samples \
                                                             for c in list(references_from_json(get_expected_file(s)))
                                                             for a in attribute_from_pop_avg(get_expected_file(s))])
def test_mp_pop_avg(sample, reference, attr):
    for section in section_from_reference(get_expected_file(sample), reference):
        compare_fields(get_expected_file(sample), output_file(sample), [reference, section, 'pop_avg', attr])


@pytest.mark.parametrize('sample,reference', [(s,c) for s in samples for c in list(references_from_json(get_expected_file(s)))])
def test_section_idx(sample, reference):
    output = output_file(sample)
    for section in section_from_reference(get_expected_file(sample), reference):
        ss, se = output[reference][section]['section_start'], output[reference][section]['section_end']
        assert 'sequence' in output[reference][section], 'ref/section/sequence is not found'
        assert len(output[reference][section][
                       'sequence']) == se - ss + 1, 'section length is not correct: {} != {}-{}+1'.format(
            len(output[reference][section]['sequence']), se, ss)


@pytest.mark.parametrize('sample,reference,section', [(ss, c, s) for ss in samples \
                                                                 for c in list(references_from_json(get_expected_file(ss))) 
                                                                 for s in section_from_reference(get_expected_file(ss), c)])
@pytest.mark.parametrize('plot', ['mutation_fraction', 'mutation_fraction_identity', 'base_coverage'])
def test_draw(sample, reference, section, plot):
    assert os.path.isfile(figure := os.path.join(top_dir, 'draw', sample, '{}__{}__{}.html'.format(reference, section,
                                                                                                   plot))), 'file {} does not exist'.format(
        figure)


if __name__ == '__main__':
    # remove test files
    test_clear_output()
    test_run_python('my_python_sample')
    test_run_cli('my_cli_sample')


from click import command
import pandas as pd

from ..cluster.bitvector import BitVector
from ..cluster.clusteringAnalysis import ClusteringAnalysis
from ..cluster.EMclustering import EMclustering
from ..util.cli import (opt_report, opt_out_dir,
                        opt_max_procs,
                        opt_max_clusters, opt_num_runs, opt_signal_thresh,
                        opt_include_gu, opt_include_del, opt_polya_max,
                        opt_min_iter, opt_max_iter, opt_convergence_cutoff, opt_min_reads, opt_verbose)

from ..util import docdef, path


params = [
    # Input/output directories
    opt_report,
    opt_out_dir,
    # Parallelization
    opt_max_procs,
    # Clustering options
    opt_max_clusters,
    opt_num_runs,
    opt_signal_thresh,
    opt_include_gu,
    opt_include_del,
    opt_polya_max,
    opt_min_iter,
    opt_max_iter,
    opt_convergence_cutoff,
    opt_min_reads
]


@command("cluster", params=params)
def cli(*args, **kwargs):
    return run(*args, **kwargs)


@docdef.auto()
def run(mp_report: tuple[str], *,
        out_dir: str,
        max_procs: int,
        # Clustering options
        max_clusters: int,
        num_runs: int,
        signal_thresh: float,
        include_gu: bool,
        include_del: bool,
        polya_max: int,
        min_iter: int,
        max_iter: int,
        convergence_cutoff: float,
        min_reads: int):
    """
    Run the clustering module. 
    """

    cluster_out_files = list()

    # Create a clustering args 
    clustering_args = dict(
        max_clusters=max_clusters,
        min_iter=min_iter,
        max_iter=max_iter,
        signal_thresh=signal_thresh,
        include_gu=include_gu,
        include_del=include_del,
        min_reads=min_reads,
        convergence_cutoff=convergence_cutoff,
        num_runs=num_runs,
        max_procs=max_procs
    )

    # Get the bitvector files in the input directory and all of its subdirectories
    for report_file in mp_report:
        report_path = path.MutVectorReportFilePath.parse(report_file)

        # Run the clustering algorithm
        bitvector = BitVector(path=report_file, signal_thresh=signal_thresh, include_gu=include_gu, min_reads=min_reads,
                              include_del=include_del)
        ca = ClusteringAnalysis(bitvector, max_clusters, num_runs, clustering_args)
        clusters = ca.run()

        # Compute the likelihood of each read for each cluster
        columns = [(k, c + 1) for k in clusters for c in range(k)]
        reads_best_cluster = pd.DataFrame(dtype=float,
                                          index=pd.Index(bitvector.read_names, name="Read Name"),
                                          columns=pd.MultiIndex.from_tuples(columns,
                                                                            names=["K", "i"]))
        for k, clusters_k in clusters.items():
            em = EMclustering(bitvector.bv, k, bitvector.read_hist, bitvector.base_to_keep, bitvector.sequence,
                              **clustering_args)
            likelihood_reads_best_cluster, _, _ = em.expectation(clusters_k[0]['mu'], clusters_k[0]['pi'])
            for c in range(k):
                reads_best_cluster.loc[:, (k, c + 1)] = likelihood_reads_best_cluster[bitvector.read_inverse, c]

        # Save the results
        cluster_out_file = str(report_path.path.with_name("clusters.csv.gz"))
        reads_best_cluster.to_csv(cluster_out_file, header=True, index=True,
                                  compression="gzip")
        cluster_out_files.append(cluster_out_file)

    return cluster_out_files



from functools import cached_property
from itertools import chain
from logging import getLogger
from pathlib import Path
from typing import Iterable

from ..util import path
from ..util.shell import run_cmd, BOWTIE2_CMD, CUTADAPT_CMD, FASTQC_CMD


logger = getLogger(__name__)


# Bowtie2 parameters
# Consider this example: Ref = ACGT, Read = AG
# Assume that we want to minimize the number of edits needed to convert
# the reference into the read sequence. The smallest number of edits is
# two, specifically these two deletions (/) from the reference: [A/G/]
# which gets a score of (2 * match - 2 * gap_open - 2 * gap_extend).
# But there are two alternative alignment, each with 3 edits:
# [Ag//] and [A//g] (substitutions marked in lowercase). Each gets the
# score (match - substitution - gap_open - 2 * gap_extend).
# In order to favor the simpler alignment with two edits,
# (2 * match - 2 * gap_open - 2 * gap_extend) must be greater than
# (match - substitution - gap_open - 2 * gap_extend); this simplifies to
# (substitution > gap_open - match). Thus, the substitution penalty and
# match bonus must be relatively large, and the gap open penalty small.
# We want to avoid introducing too many gaps, especially to prevent the
# introduction of an insertion and a deletion from scoring better than
# one substitution. Consider this example: Ref = ATAT, Read = ACTT
# The simplest alignment (the smallest number of mutations) is ActT,
# which gets a score of (2 * match - 2 * substitution).
# Another alignment with indels is A{C}T/T, where {C} means a C was
# inserted into the read and the / denotes an A deleted from the read.
# This alignment scores (3 * match - 2 * gap_open - 2 * gap_extend).
# Thus, (2 * match - 2 * substitution) must be greater than
# (3 * match - 2 * gap_open - 2 * gap_extend), which simplifies to
# (2 * gap_open + 2 * gap_extend > match + 2 * substitution).
# There are two easy solutions to these inequalities:
# - Bowtie v2.5 defaults: 6 > 5 - 2 and 2*5 + 2*3 > 2 + 2*6
# - Set every value to 1: 1 > 1 - 1 and 2*1 + 2*1 > 1 + 2*1
MATCH_BONUS = "1"
MISMATCH_PENALTY = "1,1"
N_PENALTY = "0"
REF_GAP_PENALTY = "1,1"
READ_GAP_PENALTY = "1,1"
METRICS_INTERVAL = 60  # Write metrics once every 60 seconds.


class FastqUnit(object):
    """
    Unified interface for the following sets of sequencing reads:

    - One FASTQ file of single-end reads from one sample
    - One FASTQ file of interleaved, paired-end reads from one sample
    - Two FASTQ files of mate 1 and 2 paired-end reads from one sample
    - One FASTQ file of single-end reads originating from one reference
      sequence in one sample
    - One FASTQ file of interleaved, paired-end reads originating from
      one reference sequence in one sample
    - Two FASTQ files of mate 1 and mate 2 paired-end reads originating
      from one reference sequence in one sample

    """

    MAX_PHRED_ENC = 127  # 2^7 - 1

    KEYF_SINGLE = "fastqs"
    KEYF_INTER = "fastqi"
    KEYF_MATE1 = "fastq1"
    KEYF_MATE2 = "fastq2"

    KEYD_SINGLE = "fastqs_dir"
    KEYD_INTER = "fastqi_dir"
    KEYD_MATE12 = "fastq12_dir"

    BOWTIE2_FLAGS = {KEYF_SINGLE: "-U",
                     KEYF_INTER: "--interleaved",
                     KEYF_MATE1: "-1",
                     KEYF_MATE2: "-2"}

    def __init__(self, /, *,
                 fastqs: Path | None = None,
                 fastqi: Path | None = None,
                 fastq1: Path | None = None,
                 fastq2: Path | None = None,
                 phred_enc: int,
                 one_ref: bool):
        if fastqs:
            if fastqi or fastq1 or fastq2:
                raise TypeError("Got too many FASTQ files")
            self.inputs: dict[str, Path] = {self.KEYF_SINGLE: fastqs}
            self.paired = False
            self.interleaved = False
        elif fastqi:
            if fastq1 or fastq2:
                raise TypeError("Got too many FASTQ files")
            self.inputs: dict[str, Path] = {self.KEYF_INTER: fastqi}
            self.paired = True
            self.interleaved = True
        elif fastq1:
            if not fastq2:
                raise TypeError("Got fastq1 but not fastq2")
            self.inputs: dict[str, Path] = {self.KEYF_MATE1: fastq1,
                                            self.KEYF_MATE2: fastq2}
            self.paired = True
            self.interleaved = False
        elif fastq2:
            raise TypeError("Got fastq2 but not fastq1")
        if phred_enc < 0 or phred_enc > self.MAX_PHRED_ENC:
            raise ValueError(f"Invalid Phred encoding: {phred_enc}")
        self.phred_enc = phred_enc
        self.one_ref = one_ref
        logger.debug(f"Instantiated a {self.__class__.__name__} with "
                     + ", ".join(f"{k} = {v} (type '{type(v).__name__}')"
                                 for k, v in self.inputs.items())
                     + f", phred_enc = {phred_enc}, one_ref = {one_ref}")

    @property
    def phred_arg(self):
        return f"--phred{self.phred_enc}"

    @property
    def kind(self):
        if self.paired:
            if self.interleaved:
                return "paired-end interleaved FASTQ file"
            return "paired-end separated FASTQ files"
        return "single-end FASTQ file"

    @cached_property
    def parent(self) -> Path:
        """ Return the parent directory of the FASTQ file(s). """
        parents = [inp.parent for inp in self.inputs.values()]
        if not parents:
            raise TypeError("Not parent directory")
        if any(parent != parents[0] for parent in parents[1:]):
            raise ValueError("More than one parent directory")
        return parents[0]

    @cached_property
    def paths(self) -> list[(path.OneRefReadsInFilePath |
                             path.OneRefReads1InFilePath |
                             path.OneRefReads2InFilePath |
                             path.SampleReadsInFilePath |
                             path.SampleReads1InFilePath |
                             path.SampleReads2InFilePath)]:
        """ Return a path instance for each input FASTQ file. """
        if self.one_ref:
            dtypes = {self.KEYF_SINGLE: path.OneRefReadsInFilePath,
                      self.KEYF_INTER: path.OneRefReadsInFilePath,
                      self.KEYF_MATE1: path.OneRefReads1InFilePath,
                      self.KEYF_MATE2: path.OneRefReads2InFilePath}
        else:
            dtypes = {self.KEYF_SINGLE: path.SampleReadsInFilePath,
                      self.KEYF_INTER: path.SampleReadsInFilePath,
                      self.KEYF_MATE1: path.SampleReads1InFilePath,
                      self.KEYF_MATE2: path.SampleReads2InFilePath}
        return [dtypes[key].parse(fq_path)
                for key, fq_path in self.inputs.items()]

    @cached_property
    def sample(self):
        """ Return the name of the sample of the FASTQ file(s). """
        samples: list[str] = list({fq.sample for fq in self.paths})
        if len(samples) == 0:
            raise ValueError("No sample names")
        if len(samples) > 1:
            raise ValueError(f"Sample names of input files {self} disagree: "
                             + " ≠ ".join(samples))
        return samples[0]

    @cached_property
    def ref(self):
        """ Return the name of the reference of the FASTQ file(s). """
        refs: set[str] = set()
        # Get the reference name from each FASTQ that names a reference.
        for fq in self.paths:
            try:
                refs.add(fq.ref)
            except AttributeError:
                pass
        if len(refs) == 0:
            # No FASTQ files named a reference.
            return None
        if len(refs) > 1:
            raise ValueError(f"Reference names of input files {self} disagree: "
                             + " ≠ ".join(refs))
        return list(refs)[0]

    def is_compatible_fasta(self, fasta: Path, one_ref_fasta: bool):
        """
        Return whether a given FASTA file is compatible with the FASTQ,
        which means one of the following is true:
        - The FASTA has a set of references (one_ref_fasta=False) and
          the FASTQ has reads from an entire sample.
        - The FASTA has one reference (one_ref_fasta=True) and the FASTQ
          has reads from only the same reference.
        """
        if one_ref_fasta:
            return self.ref == path.OneRefSeqInFilePath.parse(fasta).ref
        return not self.one_ref

    @property
    def cutadapt_input_args(self):
        """ Return input file arguments for Cutadapt. """
        return tuple(self.inputs.values())

    @property
    def bowtie2_inputs(self):
        """ Return input file arguments for Bowtie2. """
        return tuple(chain(*[(self.BOWTIE2_FLAGS[key], fq)
                             for key, fq in self.inputs.items()]))

    def edit(self, **fields):
        """
        Return a new FastqUnit by modifying all the FASTQ paths, using
        the given fields.

        Parameters
        ----------
        **fields: Any
            Keyword arguments to determine the new paths

        Returns
        -------
        FastqUnit
            New FastqUnit with the same keys and modified paths
        """
        return FastqUnit(**{k: fq.edit(**{**fields, path.EXT_KEY: fq.ext}).path
                            for k, fq in zip(self.inputs, self.paths,
                                             strict=True)},
                         phred_enc=self.phred_enc,
                         one_ref=self.one_ref)

    @classmethod
    def _from_files(cls, /, *,
                    key: str,
                    fq_files: Iterable[Path],
                    one_ref: bool,
                    phred_enc: int):
        """ Yield a FastqUnit for each given path to a FASTQ file of
        single-end or interleaved paired-end reads. """
        for fq in fq_files:
            if fq.suffix not in path.FQ_EXTS:
                logger.warning(f"Skipping non-FASTQ file: '{fq}'")
                continue
            try:
                yield FastqUnit(**{key: fq}, phred_enc=phred_enc, one_ref=one_ref)
            except path.PathError as error:
                logger.critical(
                    f"Failed to parse path '{fq}' as a FASTQ: {error}")

    @classmethod
    def _from_demult_files(cls, /, *,
                           key: str,
                           fq_dirs: Iterable[str],
                           phred_enc: int):
        """ Yield FastqUnits for each given directory of demultiplexed
        FASTQ files of single-end or interleaved paired-end reads. """
        for fq_dir in fq_dirs:
            yield from cls._from_files(key=key,
                                       fq_files=Path(fq_dir).iterdir(),
                                       phred_enc=phred_enc,
                                       one_ref=True)

    @classmethod
    def _from_demult_pairs(cls, /, *, fq_dirs: Iterable[str], phred_enc: int):
        """
        Yield a FastqUnit for each pair of mated, demultiplexed FASTQ files
        in a directory.

        Parameters
        ----------
        phred_enc: int
            Phred score encoding
        fq_dirs: tuple[Path]
            Directories containing the FASTQ files

        Return
        ------
        Iterable[FastqUnit]
            One for each FASTQ pair in the directory
        """
        for fq_dir in map(Path, fq_dirs):
            # Create empty dictionaries to store the FASTQ files for
            # mates 1 and 2, keyed by the name of the reference sequence
            # for each FASTQ file.
            mates1: dict[str, str] = dict()
            mates2: dict[str, str] = dict()
            # Process every FASTQ file in the directory.
            for fq_file in map(str, fq_dir.iterdir()):
                # Determine if the file is a mate 1 or 2 FASTQ file.
                # Note that using is1 = fq_file.suffix in path.FQ1_EXTS
                # would not work because each item in FQ1_EXTS includes
                # the part of the path that indicates whether the file
                # is mate 1 or 2, which comesbefore the true extension.
                # For example, the file 'mysample_R1.fq' would match the
                # item '_R1.fq' in FQ1_EXTS, but fq_file.suffix == '.fq'
                is1 = any(fq_file.endswith(ext) for ext in path.FQ1_EXTS)
                is2 = any(fq_file.endswith(ext) for ext in path.FQ2_EXTS)
                if is1 and is2:
                    # There should be no way for this error to happen,
                    # but catching it just in case.
                    logger.critical(f"FASTQ path matched both mates: {fq_file}")
                    continue
                if is1:
                    # The file name matched an extension for a mate 1
                    # FASTQ file.
                    fq_path = path.OneRefReads1InFilePath.parse(fq_file)
                    if fq_path.ref in mates1:
                        logger.critical(
                            f"Got >1 mate 1 FASTQ file for ref '{fq_path.ref}'")
                        continue
                    # Add the path to the dict of mate 1 files, keyed
                    # by reference.
                    mates1[fq_path.ref] = fq_file
                elif is2:
                    # The file name matched an extension for a mate 2
                    # FASTQ file.
                    fq_path = path.OneRefReads2InFilePath.parse(fq_file)
                    if fq_path.ref in mates2:
                        logger.critical(
                            f"Got >1 mate 2 FASTQ file for ref '{fq_path.ref}'")
                        continue
                    # Add the path to the dict of mate 2 files, keyed
                    # by reference.
                    mates2[fq_path.ref] = fq_file
                else:
                    # If a file name does not match the expected FASTQ
                    # name format, log a message but keep going, since
                    # the presence or absence of one FASTQ file will
                    # not affect the others, and this file might be an
                    # extraneous file, such as a FASTQC report or a
                    # .DS_Store file on macOS.
                    logger.warning(f"Skipping non-FASTQ file: '{fq_file}'")
            # Iterate through all references from mate 1 and mate 2 files.
            for ref in set(mates1) | set(mates2):
                fastq1 = mates1.get(ref)
                fastq2 = mates2.get(ref)
                if fastq1 is not None and fastq2 is not None:
                    # Yield a new FastqUnit from the paired FASTQ files.
                    fq_paths = {cls.KEYF_MATE1: fastq1, cls.KEYF_MATE2: fastq2}
                    yield cls(**fq_paths, phred_enc=phred_enc, one_ref=True)
                elif fastq1 is None:
                    logger.critical(f"Missing mate 1 for reference '{ref}'")
                else:
                    logger.critical(f"Missing mate 2 for reference '{ref}'")

    @classmethod
    def _from_sample_files(cls, /, *,
                           key: str,
                           fq_files: Iterable[str],
                           phred_enc: int):
        """ Yield a FastqUnit for each given path to a FASTQ file of
        single-end or interleaved paired-end reads from a sample. """
        yield from cls._from_files(key=key,
                                   fq_files=map(Path, fq_files),
                                   phred_enc=phred_enc,
                                   one_ref=False)

    @classmethod
    def _from_sample_pairs(cls, /, *,
                           fq1_files: Iterable[str],
                           fq2_files: Iterable[str],
                           phred_enc: int):
        for fq1_file, fq2_file in zip(fq1_files, fq2_files, strict=True):
            fq_paths = {cls.KEYF_MATE1: Path(fq1_file),
                        cls.KEYF_MATE2: Path(fq2_file)}
            try:
                yield FastqUnit(**fq_paths, phred_enc=phred_enc, one_ref=False)
            except (TypeError, ValueError) as error:
                logger.critical("Failed to pair up FASTQ files of mate 1 reads "
                                f"({fq1_file}) and mate 2 reads ({fq2_file}) "
                                f"due to the following error: {error}")

    @classmethod
    def from_strs(cls, /, *, phred_enc: int, **fastq_args: tuple[str, ...]):
        """
        Yield a FastqUnit for each FASTQ file (or each pair of mate 1
        and mate 2 FASTQ files) whose paths are given as strings.

        Parameters
        ----------
        phred_enc: int
            ASCII offset for encoding Phred scores
        fastq_args: tuple[str]
            FASTQ files, given as tuples of file path string:
            - fastqs: FASTQ files of single-end reads
            - fastqi: FASTQ files of interleaved paired-end reads
            - fastq1: FASTQ files of mate 1 paired-end reads; must
                      correspond 1-for-1 (in order) with fastq2
            - fastq2: FASTQ files of mate 2 paired-end reads; must
                      correspond 1-for-1 (in order) with fastq1
            - fastqs_dir: Directory of FASTQ files of single-end reads
            - fastqi_dir: Directory of FASTQ files of interleaved
                          paired-end reads
            - fastq12_dir: Directory of FASTQ files of separate mate 1
                           and mate 2 paired-end reads; for every FASTQ
                           file of mate 1 reads, there must be a FASTQ
                           file of mate 2 reads with the same sample
                           name, and vice versa.

        Yield
        -----
        FastqUnit
            FastqUnit representing the FASTQ or pair of FASTQ files.
            The order is determined primarily by the order of keyword
            arguments; within each keyword argument, by the order of
            file or directory paths; and for directories, by the order
            in which ```os.path.listdir``` returns file paths.
        """
        keys = set()
        # Directories of single-end and interleaved paired-end FASTQs
        for key in (cls.KEYD_SINGLE, cls.KEYD_INTER):
            keys.add(key)
            fq_dirs = fastq_args.get(key, ())
            yield from cls._from_demult_files(fq_dirs=fq_dirs,
                                              phred_enc=phred_enc,
                                              key=key)
        # Directories of separate mate 1 and mate 2 FASTQs
        keys.add(key := cls.KEYD_MATE12)
        fq_dirs = fastq_args.get(key, ())
        yield from cls._from_demult_pairs(fq_dirs=fq_dirs,
                                          phred_enc=phred_enc)
        # FASTQ files of single-end and interleaved paired-end reads
        for key in (cls.KEYF_SINGLE, cls.KEYF_INTER):
            keys.add(key)
            fq_files = fastq_args.get(key, ())
            yield from cls._from_sample_files(fq_files=fq_files,
                                              phred_enc=phred_enc,
                                              key=key)
        # FASTQ files of separate mate 1 and mate 2 paired-end reads
        keys.add(key := cls.KEYF_MATE1)
        fq1_files = fastq_args.get(key, ())
        keys.add(key := cls.KEYF_MATE2)
        fq2_files = fastq_args.get(key, ())
        yield from cls._from_sample_pairs(fq1_files=fq1_files,
                                          fq2_files=fq2_files,
                                          phred_enc=phred_enc)
        if extras := set(fastq_args) - keys:
            # This should never happen; catching just in case.
            raise TypeError(f"Got extra keyword arguments: {extras}")

    def __str__(self):
        return f"{self.kind} {' and '.join(map(str, self.inputs.values()))}"


def run_fastqc(fq_unit: FastqUnit, out_dir: Path, extract: bool):
    """ Run FASTQC on the given FASTQ unit. """
    logger.info(f"Began FASTQC of {fq_unit}")
    # FASTQC command, including whether to extract automatically
    cmd = [FASTQC_CMD, "--extract" if extract else "--noextract"]
    # FASTQC output directory
    out_dir.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensured directory: {out_dir}")
    cmd.extend(["-o", out_dir])
    # Input FASTQ files
    cmd.extend(fq_unit.inputs.values())
    # Run FASTQC
    run_cmd(cmd)
    logger.info(f"Ended FASTQC of {fq_unit}")


def run_cutadapt(fq_inp: FastqUnit,
                 fq_out: FastqUnit, *,
                 n_procs: int,
                 cut_q1: int,
                 cut_q2: int,
                 cut_g1: str,
                 cut_a1: str,
                 cut_g2: str,
                 cut_a2: str,
                 cut_o: int,
                 cut_e: float,
                 cut_indels: bool,
                 cut_nextseq: bool,
                 cut_discard_trimmed: bool,
                 cut_discard_untrimmed: bool,
                 cut_m: int):
    """ Trim adapters and low-quality bases with Cutadapt. """
    logger.info(f"Began trimming {fq_inp}")
    # Cutadapt command
    cmd = [CUTADAPT_CMD, "--cores", n_procs]
    # Quality trimming
    if cut_nextseq:
        cut_qnext = max(cut_q1, cut_q2)
        if cut_q1 != cut_q2:
            logger.warning("NextSeq trimming takes one quality level, but got "
                           f"two ({cut_q1} and {cut_q2}); using {cut_qnext}")
        cmd.extend(["--nextseq-trim", cut_qnext])
    else:
        cmd.extend(["-q", cut_q1, "-Q", cut_q2])
    # Adapter trimming
    adapters = {"g": cut_g1, "a": cut_a1,
                "G": cut_g2, "A": cut_a2}
    for arg, adapter in adapters.items():
        if adapter and (fq_inp.paired or arg.islower()):
            for adapt in adapter:
                cmd.extend([f"-{arg}", adapt])
    cmd.extend(["-O", cut_o])
    cmd.extend(["-e", cut_e])
    cmd.extend(["-m", cut_m])
    if not cut_indels:
        cmd.append("--no-indels")
    if cut_discard_trimmed:
        cmd.append("--discard-trimmed")
    if cut_discard_untrimmed:
        cmd.append("--discard-untrimmed")
    # FASTQ format
    if fq_inp.interleaved:
        cmd.append("--interleaved")
    # Output files
    for flag, value in zip(("-o", "-p"), fq_out.paths, strict=False):
        cmd.extend([flag, value])
    # Input files
    cmd.extend(fq_inp.cutadapt_input_args)
    # Make the output directory.
    fq_out.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensured directory: {fq_out.parent}")
    # Run Cutadapt.
    run_cmd(cmd)
    logger.info(f"Ended trimming {fq_inp}; output {fq_out}")


def run_bowtie2(fq_inp: FastqUnit,
                index_pfx: Path,
                sam_out: Path, *,
                n_procs: int,
                bt2_local: bool,
                bt2_discordant: bool,
                bt2_mixed: bool,
                bt2_dovetail: bool,
                bt2_contain: bool,
                bt2_unal: bool,
                bt2_score_min: str,
                bt2_i: int,
                bt2_x: int,
                bt2_gbar: int,
                bt2_l: int,
                bt2_s: str,
                bt2_d: int,
                bt2_r: int,
                bt2_dpad: int,
                bt2_orient: str):
    """ Align reads to the reference with Bowtie 2. """
    logger.info(f"Began aligning {fq_inp} to {index_pfx}")
    # Bowtie2 command
    cmd = [BOWTIE2_CMD]
    # Resources
    cmd.extend(["--threads", n_procs])
    # Alignment
    if bt2_local:
        cmd.append("--local")
    cmd.extend(["--gbar", bt2_gbar])
    cmd.extend(["--dpad", bt2_dpad])
    cmd.extend(["-L", bt2_l])
    cmd.extend(["-i", bt2_s])
    cmd.extend(["-D", bt2_d])
    cmd.extend(["-R", bt2_r])
    # Scoring
    cmd.append(fq_inp.phred_arg)
    cmd.append("--ignore-quals")
    cmd.extend(["--ma", MATCH_BONUS])
    cmd.extend(["--mp", MISMATCH_PENALTY])
    cmd.extend(["--np", N_PENALTY])
    cmd.extend(["--rfg", REF_GAP_PENALTY])
    cmd.extend(["--rdg", READ_GAP_PENALTY])
    # Filtering
    if not bt2_unal:
        cmd.append("--no-unal")
    cmd.extend(["--score-min", bt2_score_min])
    cmd.extend(["-I", bt2_i])
    cmd.extend(["-X", bt2_x])
    # Mate pair orientation
    cmd.append(f"--{bt2_orient}")
    if not bt2_discordant:
        cmd.append("--no-discordant")
    if not bt2_contain:
        cmd.append("--no-contain")
    if bt2_dovetail:
        cmd.append("--dovetail")
    if not bt2_mixed:
        cmd.append("--no-mixed")
    # Formatting
    cmd.append("--xeq")
    # Metrics
    cmd.extend(["--met-stderr", "--met", METRICS_INTERVAL])
    # Input and output files
    cmd.extend(["-S", sam_out])
    cmd.extend(["-x", index_pfx])
    cmd.extend(fq_inp.bowtie2_inputs)
    # Make the output directory.
    sam_out.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensured directory: {sam_out.parent}")
    # Run alignment.
    run_cmd(cmd)
    logger.info(f"Ended aligning {fq_inp}; output {sam_out}")



from functools import wraps
from logging import getLogger
from os import makedirs, path, rmdir
from typing import Callable


logger = getLogger(__name__)

# DREEM lock directory (for function lock_output)
LOCK_DIR = ".dreem-lock"


def lock_output(run: Callable):
    """ Prevent multiple instances of DREEM from using the same output
    and temporary directories simultaneously, to avoid data races. """
    @wraps(run)
    def with_locked_output(*args, out_dir, temp_dir, **kwargs):
        # Create hidden lock directories inside output and temp that
        # exist only for the duration of the run function.
        out_lock = path.join(out_dir, LOCK_DIR)
        temp_lock = path.join(temp_dir, LOCK_DIR)
        all_locks = [out_lock, temp_lock]
        my_locks = list()
        try:
            for lock in all_locks:
                try:
                    # Creating the locks will fail if another instance of
                    # DREEM (with its own locks) is currently running.
                    makedirs(lock, exist_ok=False)
                except FileExistsError:
                    # Quit because another instance of DREEM is running.
                    raise SystemExit(f"An instance of DREEM with directory "
                                     f"{path.dirname(lock)} is running, "
                                     f"which can cause data races.")
                else:
                    my_locks.append(lock)
                    logger.debug(f"Created directory lock: {lock}")
            try:
                # Call the DREEM run function and capture its return value.
                result = run(*args, out_dir=out_dir, temp_dir=temp_dir, **kwargs)
            except Exception:
                raise
            else:
                return result
        finally:
            # Always delete the locks after the run finishes, whether
            # normally or with an error.
            while my_locks:
                lock = my_locks.pop()
                try:
                    rmdir(lock)
                    logger.debug(f"Removed directory lock: {lock}")
                except FileNotFoundError:
                    logger.error(f"Failed to delete lock {lock}; please delete "
                                 f"this directory yourself, if it exists")
    # Return the decorator.
    return with_locked_output


def get_num_parallel(n_tasks: int,
                     max_procs: int,
                     parallel: bool,
                     hybrid: bool = False) -> tuple[int, int]:
    """ Determine how to parallelize the tasks.

    Parameters
    ----------
    n_tasks: int (≥ 1)
        Number of tasks to parallelize
    max_procs: int (≥ 1)
        Maximum number of processes to run at one time
    parallel: bool
        Whether to permit multiple tasks to be run in parallel
    hybrid: bool (default: False)
        Whether to allow both multiple tasks to run in parallel and,
        at the same, each task to run multiple processes in parallel

    Returns
    -------
    int (≥ 1)
        Number of tasks to run in parallel
    int (≥ 1)
        Number of processes to run for each task
    """
    if n_tasks >= 1 and max_procs >= 1:
        # This function only works if there is at least one task to
        # parallelize, at least one process is allowed, and parallel
        # is a valid option.
        if parallel:
            # Multiple tasks may be run in parallel. The number of tasks
            # run in parallel cannot exceed 1) the total number of tasks
            # and 2) the user-specified maximum number of processes.
            n_tasks_parallel = min(n_tasks, max_procs)
        else:
            # Otherwise, only one task at a time can be run.
            n_tasks_parallel = 1
        if n_tasks_parallel == 1 or hybrid:
            # Each individual task can be run by multiple processes in
            # parallel, as long as either 1) multiple tasks are not run
            # simultaneously in parallel (i.e. n_tasks_parallel == 1)
            # or 2) the calling function sets hybrid=True, which lets
            # multiple tasks run in parallel and each run with multiple
            # processes. Only the alignment module can simultaneously
            # run multiple tasks and multiple processes for each task
            # because its two most computation-heavy processes (cutadapt
            # and bowtie2) come with their own parallelization abilities
            # that can work independently of Python's multiprocessing
            # module. However, the other modules (e.g. vectoring) are
            # parallelized using the multiprocessing module, which does
            # not support "nesting" parallelization in multiple layers.
            # Because n_tasks_parallel is either 1 or the smallest of
            # n_tasks and n_procs (both of which are ≥ 1), it must be
            # that 1 ≤ n_tasks_parallel ≤ n_procs, and therefore that
            # 1 ≤ n_procs / n_tasks_parallel ≤ n_procs, so the
            # integer quotient must be a valid number of processes.
            n_procs_per_task = max_procs // n_tasks_parallel
        else:
            # Otherwise, only one process can work on each task.
            n_procs_per_task = 1
    else:
        logger.warning("Defaulting to 1 process due to invalid number of "
                       f"tasks ({n_tasks}) and/or processes ({max_procs}).")
        n_tasks_parallel = 1
        n_procs_per_task = 1
    return n_tasks_parallel, n_procs_per_task



import json
import os
import sys


# Set environment variable DATAPATH to RNAstructure data tables.
data_path_key = "DATAPATH"
if os.environ.get(data_path_key) is None:
    rs_dir = "rnastructure"
    dt_dir = "data_tables"
    env_dir = os.path.dirname(os.path.dirname(sys.executable))
    for root, dirs, files in os.walk(env_dir):  # FIXME: this might not be portable
        if (os.path.basename(root) == dt_dir
                and os.path.basename(os.path.dirname(root)) == rs_dir):
            # RNAstructure data tables were found in root.
            # Set DATAPATH environment variable to root.
            os.environ[data_path_key] = root
            break
    else:
        # RNAstructure data tables were not found.
        #raise FileNotFoundError(os.path.join(rs_dir, dt_dir))
        pass


def run_command(cmd):
    import subprocess
    print(cmd)
    process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    return output.decode('utf-8')
                

class RNAstructure(): 
    def __init__(   self, 
                    rnastructure_path='', 
                    temp = 'temp/rnastructure/', 
                    rnastructure_use_temp = None,  
                    rnastucture_fold_args = None,
                    rnastructure_use_dms = None,
                    rnastructure_dms_min_unpaired_value = None,
                    rnastructure_dms_max_paired_value = None,
                    rnastructure_deltag_ensemble = None,
                    rnastructure_probability = None,
                    verbose = False,
                 ) -> None:
        self.rnastructure_path = rnastructure_path
        self.rnastructure_use_temp = rnastructure_use_temp
        self.rnastucture_fold_args = rnastucture_fold_args
        self.rnastructure_use_dms = rnastructure_use_dms
        self.rnastructure_dms_min_unpaired_value = rnastructure_dms_min_unpaired_value
        self.rnastructure_dms_max_paired_value = rnastructure_dms_max_paired_value
        self.rnastructure_deltag_ensemble = rnastructure_deltag_ensemble
        self.rnastructure_probability = rnastructure_probability
        self.verbose = verbose
        
        if self.rnastructure_use_dms != None:
            assert self.rnastructure_dms_min_unpaired_value != None, "If you want to use DMS, you need to specify the min and max values for the DMS data"
            assert self.rnastructure_dms_max_paired_value != None, "If you want to use DMS, you need to specify the min and max values for the DMS data"
            
        if len(self.rnastructure_path) > 0:
            if self.rnastructure_path[-1] != '/':
                self.rnastructure_path += '/'
        self.directory = temp
        if os.path.exists(os.path.join(self.directory, 'ledger.json')):
            self.ledger = json.load(open(os.path.join(self.directory, 'ledger.json'), 'r'))
        else:
            self.ledger = {}

    def __make_files(self, temp_prefix='temp'):
        self.pfs_file = os.path.join(self.directory, temp_prefix+'.pfs')
        self.ct_file = os.path.join(self.directory, temp_prefix+'.ct')
        self.dot_file = os.path.join(self.directory, temp_prefix+'.dot')
        self.fasta_file = os.path.join(self.directory, temp_prefix+'.fasta')
        self.prob_file = os.path.join(self.directory, temp_prefix+'_prob.txt')

    def fit(self, sequence, reference='reference'):
        self.sequence = sequence
        self.__make_temp_folder()
        self.__make_files()
        self.__create_fasta_file(reference, sequence)

    def predict_ensemble_energy(self):
        cmd = f"{self.rnastructure_path}EnsembleEnergy {self.fasta_file} --sequence"
        splitted_output = run_command(cmd).split(' ')
        return float(splitted_output[splitted_output.index(f"kcal/mol\n\nEnsemble")-1])

    def predict_partition(self, temperature_k =None):
        cmd = f"{self.rnastructure_path}partition {self.fasta_file} {self.pfs_file}"
        if temperature_k != None:
            cmd += ' --temperature '+str(temperature_k)
        run_command(cmd)
        run_command(self.rnastructure_path+'ProbabilityPlot '+ self.pfs_file + ' -t '+self.prob_file)
        with open(self.prob_file,"r") as f:
            lines=f.readlines()
            out={'i':[],'j':[],'p':[]}
            for x in range(len(lines)):
                if x>1:
                    ls=lines[x].split("\t")
                    out["i"]+=[int(ls[0])]
                    out["j"]+=[int(ls[1])]
                    out["p"]+=[float(ls[2])]
        return self.__cast_pairing_prob(out)

    def predict_reference_deltaG(self):
        cmd = f"{self.rnastructure_path}Fold {self.fasta_file} {self.ct_file}"
        run_command(cmd)
        assert os.path.exists(self.ct_file), f"{self.ct_file} does not exist, check that RNAstructure works. Command: {cmd}"
        assert os.path.getsize(self.ct_file) != 0, f"{self.ct_file} is empty, check that RNAstructure works. Command: {cmd}"
        return self.__extract_deltaG_struct()

    def __make_temp_folder(self):
        isExist = os.path.exists(self.directory)
        if not isExist:
            os.makedirs(self.directory)
        return self.directory

    def __create_fasta_file(self, reference, sequence):
        # push the ref into a temp file
        temp_fasta = open(self.fasta_file, 'w')
        temp_fasta.write('>'+reference+'\n'+sequence)
        temp_fasta.close()

    # cast the temp file into a dot_bracket structure and extract the attributes
    def __extract_deltaG_struct(self):
        run_command(f"{self.rnastructure_path}ct2dot {self.ct_file} 1 {self.dot_file}")
        temp_dot = open(self.dot_file, 'r')
        first_line = temp_dot.readline().split()
        # If only dots in the structure, no deltaG 

        if len(first_line) == 4:
            _, _, deltaG, _ = first_line
            deltaG = float(deltaG)
        if len(first_line) == 1:
            deltaG, _ = 0.0, first_line[0][1:]

        sequence = temp_dot.readline()[:-1] #  Remove the \n
        structure = temp_dot.readline()[:-1] # Remove the \n
        return deltaG, structure
    
    def dump_ledger(self):
        json.dump(self.ledger, open(os.path.join(self.directory, 'ledger.json'), 'w'))

    def run(self, sequence):
        if sequence in self.ledger:
            deltaG, structure = self.ledger[sequence]['deltaG'], self.ledger[sequence]['structure']
        else:
            self.fit(sequence)
            deltaG, structure = self.predict_reference_deltaG()
            self.ledger[sequence] = {'deltaG':deltaG, 'structure':structure}
            for file in os.listdir(self.directory):
                if file.startswith('temp'):
                    os.remove(os.path.join(self.directory, file))
        if deltaG == 'void': 
            deltaG = 0.0
        else:
            deltaG = float(deltaG)
        return {'deltaG':deltaG, 'structure':structure}

if __name__ == "__main__":
    rna = RNAstructure('/Users/ymdt/src/RNAstructure/exe/', temp = 'my_temp')
    rna.fit(sequence='AAGATATTCGAAACCACTCGATCGACTAGCATCAGCTGACTAGCTAGCATGCATCAAGAATATCTT')
    print("DeltaG + structure:", rna.predict_reference_deltaG())
    print("Ens. energy:", rna.predict_ensemble_energy())
    print("One line command:", rna.run('AAGATATTCGAAACCACTCGATCGACTAGCATCAGCTGACTAGCTAGCATGCATCAAGAATATCTT'))
    rna.dump_ledger()
